/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "app/common",
    "app/messages",
    "app/wizardData",
    "aps/load",
    "aps/Memory",
    "aps/ResourceStore",
    "aps/xhr",
    "dijit/registry",
    "dojo/promise/all",
    "dojox/mvc/at",
    "dojox/mvc/getPlainValue",
    "dojox/mvc/getStateful"
], function (
    common,
    messages,
    wd,
    load,
    Memory,
    Store,
    xhr,
    registry,
    all,
    at,
    getPlainValue,
    getStateful
) {
    var globalId = aps.context.vars.globals.aps.id,
        firstStepViewId = common.getRelatedViewId('new-select'),
        secondStepViewId = common.getRelatedViewId('new-configure');

    function serviceUserExist(login, serviceUsers) {
        return serviceUsers.some(function (serviceUser) {
            return serviceUser.login === login;
        });
    }

    function serviceUserHasBackupAccount(serviceUser, backupUsers) {
        return backupUsers.some(function (backupUser) {
            return backupUser.serviceUserId === serviceUser.aps.id;
        });
    }

    function prepareServiceUsers(serviceUsers, backupUsers) {
        var users = [];
        serviceUsers.forEach(function (serviceUser) {
            if (!serviceUserHasBackupAccount(serviceUser, backupUsers)) {
                users.push({
                    id: serviceUser.aps.id,
                    name: "{0} ({1})".replace('{0}', common.escapeAngularBrackets(serviceUser.displayName)).replace('{1}', serviceUser.login)
                });
            }
        });
        return users;
    }

    xhr.get("/aps/2/resources/" + globalId + "/settings").then(function (settings) {
        var serviceUserStore = new Store({
                apsType: !settings.ccpv1_show_staff_members ?
                    "http://aps-standard.org/types/core/service-user/1.0" :
                    "http://parallels.com/aps/types/pa/user/1.0",
                target: "/aps/2/resources",
                baseQuery: 'sort(displayName)'
            }),
            backupUserStore = new Store({
                apsType: common.TYPES.USERS,
                target: "/aps/2/resources"
            });

        all({
            backupUsers: backupUserStore.query(),
            serviceUsers: serviceUserStore.query()
        }).then(function (response) {
            var firstStepData = wd.getWizardData(firstStepViewId),
                secondStepData = wd.getWizardData(secondStepViewId),
                existingUserModel = getStateful(
                    firstStepData.existingUser ? firstStepData.existingUser : {aps: {}}
                ),
                newUserModel = getStateful(
                    firstStepData.newUser ? firstStepData.newUser : {aps: {type: "http://aps-standard.org/types/core/service-user/1.0"}}
                ),
                userSwitchModel = getStateful(
                    firstStepData.userSwitch ? firstStepData.userSwitch : {existing: true}
                ),
                properServiceUsers = prepareServiceUsers(response.serviceUsers, response.backupUsers),
                userStore = new Memory({
                    data: properServiceUsers,
                    idProperty: 'id'
                });

            if (!properServiceUsers.length) {
                userSwitchModel.set({existing: false});
            }
            else {
                if (!existingUserModel.aps.id) {
                    existingUserModel.aps.id = properServiceUsers[0].id;
                }
            }

            load(["aps/PageContainer", {id: "apsPageContainer"}, [
                ["aps/WizardControl", {
                    id: common.genId('wizardcontrol'),
                    steps: [
                        {viewId: firstStepViewId, text: _("Select user"), active: true},
                        {viewId: secondStepViewId, text: _("Configure service")}
                    ]
                }],
                ["aps/FieldSet", {id: common.genId('fieldset'), showLabels: false}, [
                    ["aps/RadioButton", {
                        id: "existingUser",
                        description: _("Existing user"),
                        checked: at(userSwitchModel, "existing").transform({
                            format: function (value) {
                                return value;
                            },
                            parse: function (value) {
                                return value;
                            }
                        }),
                        disabled: !properServiceUsers.length,
                        hint: (properServiceUsers.length ? null : _("There are no users without the Cyber Protection service."))
                    }],
                    ["aps/RadioButton", {
                        id: "newUser",
                        description: _("Create new user"),
                        checked: at(userSwitchModel, "existing").transform({
                            format: function (value) {
                                return !value;
                            },
                            parse: function (value) {
                                return !value;
                            }
                        })
                    }]
                ]],
                ["aps/FieldSet", {
                    id: "selectFieldSet",
                    title: _("Select user"),
                    visible: at(userSwitchModel, "existing").transform({
                        format: function (value) {
                            return value;
                        },
                        parse: function (value) {
                            return value;
                        }
                    })
                }, [
                    ["aps/Select", {
                        id: "user",
                        store: userStore,
                        value: at(existingUserModel.aps, "id"),
                        labelAttr: "name",
                        title: _("User"),
                        required: true
                    }]
                ]],
                ["aps/FieldSet", {
                    id: "createFieldSet",
                    title: _("Create user"),
                    visible: at(userSwitchModel, "existing").transform({
                        format: function (value) {
                            return !value;
                        },
                        parse: function (value) {
                            return !value;
                        }
                    })
                }, [
                    ["aps/TextBox", {
                        id: "displayName",
                        label: _("Display name"),
                        pattern: '^\\s*([^\\s]+)(\\s+[^\\s]+)*\\s*$',
                        value: at(newUserModel, "displayName").transform({
                            parse: function (value) {
                                value = value.trim();
                                if (value === "") {
                                    newUserModel.set({
                                        givenName: "",
                                        familyName: "",
                                        fullName: ""
                                    });
                                    return value;
                                }

                                var regex = /([^\s]+)/g, parts = [], matches,
                                    givenName = "", familyName = "", fullName = "";

                                while (matches = regex.exec(value)) {
                                    parts.push(matches[1]);
                                }

                                if (parts.length > 0) {
                                    givenName = parts[0];
                                    fullName = parts[0];
                                }

                                if (parts.length > 1) {
                                    var index = parts.length - 1;
                                    familyName = parts[index];
                                    fullName += " " + parts[index];
                                }

                                newUserModel.set({
                                    givenName: givenName,
                                    familyName: familyName,
                                    fullName: fullName
                                });

                                return value;
                            }
                        }),
                        required: true
                    }],
                    ["aps/TextBox", {
                        id: "login",
                        label: _("Login"),
                        hint: _("Login should be in the e-mail address format e.g. user@example.com."),
                        value: at(newUserModel, "login"),
                        pattern: '^.+@.+\\..+$',
                        required: true,
                        invalidMessage: _("Invalid format of login. Login should be in the e-mail address format.")
                    }],
                    ["./js/widgets/Password.js", {
                        id: "password",
                        label: _("Password"),
                        hint: _("7-characters minimum, including letters in different cases and digits"),
                        value: at(newUserModel, "password"),
                        required: true
                    }]
                ]]
            ]]).then(function () {
                aps.app.onNext = function () {
                    messages.removeAll();
                    var fieldSet = userSwitchModel.existing ? registry.byId("selectFieldSet") : registry.byId("createFieldSet");
                    if (!fieldSet.validate()) {
                        messages.displayMessage(_("Incorrect or empty field value(s). Please retype info in the highlighted fields."), "error");
                        aps.apsc.cancelProcessing();
                        return;
                    }

                    if (!userSwitchModel.existing) {
                        var login = registry.byId("login").get("value");
                        if (serviceUserExist(login, response.serviceUsers)) {
                            messages.displayMessage(_("User with login '{0}' already exists.").replace("{0}", login), "error");
                            aps.apsc.cancelProcessing();
                            return;
                        }
                    }

                    firstStepData.existingUser = getPlainValue(existingUserModel);
                    firstStepData.newUser = getPlainValue(newUserModel);
                    firstStepData.userSwitch = getPlainValue(userSwitchModel);
                    wd.setWizardData(firstStepData, firstStepViewId);

                    secondStepData.user = firstStepData.userSwitch.existing ? firstStepData.existingUser : firstStepData.newUser;
                    wd.setWizardData(secondStepData, secondStepViewId);

                    aps.apsc.gotoView(secondStepViewId);
                };
            });
        }).otherwise(messages.displayError);
    }, messages.displayError);

    aps.app.onCancel = function () {
        wd.resetWizardData(firstStepViewId);
        wd.resetWizardData(secondStepViewId);
        aps.apsc.gotoView(common.getRelatedViewId('list'));
    };
});