/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/_base/declare",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "aps/Button",
    "aps/Container",
    "aps/DropDownButton",
    "aps/Grid",
    "aps/Memory",
    "aps/Toolbar",
    "./base/Status",
    "./../components/getStorageOutput",
    "./../components/BackupAdminActions",
    "./../../common",
    "./../../messages"
], function (
    declare,
    at,
    getStateful,
    Button,
    Container,
    DropDownButton,
    Grid,
    Memory,
    Toolbar,
    Status,
    getStorageOutput,
    BackupAdminActions,
    common,
    messages
) {

    var statusInfo = {
        updating: {"label": _("UPDATING"), "type": "default", isLoad: true},
        deleting: {"label": _("DELETING"), "type": "danger", isLoad: true},
        notAssigned: {"label": _("NOT ASSIGNED"), "type": "default", isLoad: false},
        notActivated: {"label": _("NOT ACTIVATED"), "type": "warning", isLoad: false},
        activating: {"label": _("ACTIVATING"), "type": "warning", isLoad: true, useIcon: false},
        activated: {"label": _("ACTIVATED"), "type": "success", isLoad: false},
        disabled: {"label": _("DISABLED"), "type": "default", isLoad: true}
    };

    function createStatus(row) {
        var status = new Status({
            useIcon: false,
            gridSize: 'md-12',
            status: at(row.statusModel, "name"),
            statusInfo: statusInfo
        });
        status.startup();

        return status;
    }

    function addStatefulStatus(data) {
        data.forEach(function (item) {
            item.statusModel = item.statusModel || new getStateful({"name": item.backupUserStatus});
        });
        return data;
    }

    function prepareAllowedDevicesLimit(statistic, tenant) {

        function getCounterLimit(counter) {
            if (!(counter && counter.limit)) {
                return null;
            }
            if (counter.limit < 0) {
                return _("Unlimited");
            }
            return String(counter.limit);
        }

        function checkUnlimited(counter, tenantCounter) {
            return !tenantCounter || !counter || counter.limit < 0;
        }

        let counters = common.getAvailableCountersForArea("quota_list_view_ccpv2"),
            isUnlimited = true;

        counters.forEach(function (counter) {
            if (statistic[counter.name]) {
                isUnlimited = isUnlimited && checkUnlimited(statistic[counter.name], tenant[counter.name]);
            }
        });

        if (isUnlimited) {
            return _("Unlimited");
        }

        var result = [];
        counters.forEach(function (counter) {
            if (!statistic[counter.name]) {
                return;
            }
            let limit = getCounterLimit(statistic[counter.name]);
            if (limit) {
                result.push(_("{0} " + counter.label + "(s)").replace("{0}", limit));
            }
        });

        if (!result.length) {
            return _('None');
        }
        return result.join(', ');
    }

    return declare(Container, {
        title: _("Service Users"),
        users: [],
        gridSize: "md-12 xs-12",

        _setUsersAttr: function (val) {
            var tenant = aps.context.vars.tenant,
                userQuota = common.getUserQuota(tenant.profile),
                userStore = new Memory({data: addStatefulStatus(val), idProperty: "aps.id"}),
                layout = [
                    {
                        name: _('User'),
                        field: "displayName",
                        type: "resourceName",
                        filter: true
                    },
                    {
                        name: _("Service Status"),
                        sortable: false,
                        field: "backupUserStatus",
                        renderCell: createStatus
                    },
                    {
                        name: _("Cloud Storage usage"),
                        field: "backupUserStatistic",
                        sortable: false,
                        renderCell: function (row, data) {
                            let storageData = null,
                                storageCounters = common.getAvailableCountersForArea("storage_usage_only");
                            data && storageCounters.forEach(function (counter) {
                                if (data.hasOwnProperty(counter.name)) {
                                    storageData = data[counter.name];
                                }
                            });

                            if (!storageData || storageData.usage === null) {
                                return _('N/A');
                            }

                            return getStorageOutput({
                                limitedLegend: _("${usage} ${usageUnit} of ${limit} ${limitUnit} used"),
                                unlimitedLegend: _("${usage} ${usageUnit} used"),
                                usage: storageData.usage || 0,
                                limit: storageData.limit || 0
                            });
                        }
                    },
                    {
                        name: '',
                        field: "backupUser",
                        sortable: false,
                        renderCell: function (row) {
                            var backupAccountActions = new BackupAdminActions(row, tenant),
                                separatorMenuItem = {
                                    type: "separator"
                                },
                                viewMenuItem = {
                                    iconClass: "fa-user",
                                    label: _("View User"),
                                    onClick: function () {
                                        aps.apsc.gotoView("http://www.parallels.com/ccp-users#viewUser", row.aps.id);
                                    }
                                },
                                assignMenuItem = {
                                    iconClass: "fa-check",
                                    label: _("Assign Service"),
                                    onClick: function () {
                                        backupAccountActions.assign().then(null, messages.displayError);
                                    }
                                },
                                activateMenuItem = {
                                    iconClass: "fa-magic",
                                    type: "warning",
                                    label: _("Activate"),
                                    onClick: function () {
                                        backupAccountActions.activate().then(null, messages.displayError);
                                    }
                                },
                                configureMenuItem = {
                                    iconClass: "fa-pencil",
                                    label: _("Change Settings"),
                                    onClick: function () {
                                        backupAccountActions.configure().then(null, messages.displayError);
                                    }
                                },
                                unassignMenuItem = {
                                    iconClass: "fa-times",
                                    label: _("Unassign Service"),
                                    onClick: function () {
                                        backupAccountActions.unassign().then(null, messages.displayError);
                                    }
                                },
                                enableMenuItem = {
                                    iconClass: "fa-times",
                                    label: _("Enable"),
                                    onClick: function () {
                                        backupAccountActions.enable().then(function (userData) {
                                            if (userData.userstatus === common.STATUS.ENABLED) {
                                                row.statusModel.set("name", "activated");
                                                grid.refresh();
                                            }
                                        }, messages.displayError);
                                    }
                                },
                                disableMenuItem = {
                                    iconClass: "fa-times",
                                    label: _("Disable"),
                                    onClick: function () {
                                        backupAccountActions.disable().then(function (userData) {
                                            if (userData.userstatus === common.STATUS.DISABLED) {
                                                row.statusModel.set("name", "disabled");
                                                grid.refresh();
                                            }
                                        }, messages.displayError);
                                    }
                                },
                                loginToImpersonatedBackupConsole = {
                                    iconClass: "fa-external-link",
                                    label: _("Log into Backup Console"),
                                    onClick: function () {
                                        var w = window.open('', '_blank');
                                        backupAccountActions.loginToConsole().then(function (uri) {
                                            w.location.href = uri;
                                        }, function (error) {
                                            w.close();
                                            messages.displayError(error);
                                        });
                                    }
                                },
                                menuItems = [];

                            switch (row.statusModel.name) {
                                case 'notAssigned':
                                    menuItems = [assignMenuItem, viewMenuItem];
                                    break;

                                case 'notActivated':
                                    menuItems = [activateMenuItem, viewMenuItem, separatorMenuItem, unassignMenuItem];
                                    break;

                                case 'activated':
                                    menuItems = [loginToImpersonatedBackupConsole, separatorMenuItem, configureMenuItem, viewMenuItem];
                                    if (row.backupUser.servicestatus == common.STATUS.ENABLED) {
                                        menuItems.push(separatorMenuItem);
                                        menuItems.push(disableMenuItem);
                                    }
                                    menuItems.push(unassignMenuItem);
                                    break;
                                case 'disabled':
                                    configureMenuItem.disabled = true;
                                    menuItems = [configureMenuItem, viewMenuItem, separatorMenuItem, enableMenuItem, unassignMenuItem];
                                    break;
                                default:
                                    configureMenuItem.disabled = true;
                                    unassignMenuItem.disabled = true;
                                    menuItems = [configureMenuItem, viewMenuItem, separatorMenuItem, unassignMenuItem];
                            }

                            return new DropDownButton({
                                iconClass: "fa-cog",
                                label: _("Actions"),
                                items: menuItems
                            });
                        }
                    }
                ];

            var toolbar = new Toolbar(),
                grid = new Grid({
                    selectionMode: "multiple",
                    style: "padding: 0px;",
                    columns: layout,
                    store: userStore,
                    showPaging: true,
                    rowsPerPage: 20,
                    pageSizeOptions: [20],
                    apsResourceViewId: 'http://www.parallels.com/ccp-users#viewUser'
                }),
                addButton = new Button({
                    autoBusy: false,
                    label: _("Add New Users"),
                    iconClass: "fa-plus",
                    isDefault: true,
                    onClick: function () {
                        aps.apsc.gotoView("http://www.parallels.com/ccp-users#addUser");
                    }
                }),
                assignButton = new Button({
                    autoBusy: false,
                    label: _("Assign Service"),
                    iconClass: "fa-pencil",
                    isDefault: true,
                    disabled: true,
                    onClick: function () {
                        var usersIds = grid.selectionArray,
                            selectedUsersIds = [];
                        usersIds.forEach(function (userId) {
                            var user = grid.store.get(userId);
                            if (!user.backupUser) {
                                selectedUsersIds.push(userId);
                            }
                        });
                        aps.apsc.gotoView("nextcp-ccp-multiple-setquotas", null, {
                            usersIds: selectedUsersIds
                        });
                    }
                });

            grid.selectionArray.watchElements(function () {
                var users = grid.selectionArray.filter(function (userId) {
                    var user = grid.store.get(userId);
                    return !user.backupUser;
                });
                assignButton.set("disabled", !users.length);
            });

            toolbar.addChild(addButton);
            toolbar.addChild(assignButton);
            grid.addChild(toolbar);

            this.removeAll();
            this.addChild(grid);

            this.users = val;
        }
    });
});



