/*
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/Deferred",
    "dojo/promise/all",
    "aps/xhr",
    "app/common",
    "app/tags"
], function (
    Deferred,
    all,
    xhr,
    common,
    tags
) {
    /**
     * Returns filtered storages array.
     * @param storages
     * @param profile
     * @returns {Array}
     */
    function filterStorages(storages, profile) {
        var profileStorages = [];
        if (profile.storages) {
            profileStorages = storages.filter(function (storage) {
                return profile.storages.some(function (profileStorage) {
                    return storage.storageId == profileStorage.storageId &&
                        storage.datacenterId == profileStorage.datacenterId;
                });
            });
        }
        return profileStorages;
    }

    /**
     * Returns filtered datacenters array.
     * @param datacenters
     * @param profile
     * @returns {Array}
     */
    function filterDatacenters(datacenters, profile) {
        var profileDatacenters = [];
        if (profile.datacenters) {
            profileDatacenters = datacenters.filter(function (datacenter) {
                return profile.datacenters.some(function (profileDatacenter) {
                    return datacenter.id == profileDatacenter.datacenterId;
                });
            });
        }
        return profileDatacenters;
    }

    /**
     * Returns storages formatted list.
     * @param profileStorages
     * @param profile
     * @returns {*}
     */
    function makeStoragesOutput(profileStorages, profile) {
        if (profile && profile.storagePolicy == common.STORAGE_POLICY.OWN) {
            return _("Resellers must use their own storage");
        }
        return !!profileStorages.length ? tags.ul(profileStorages.map(function (storage) {
                return _("__STORAGE__ (__DATACENTER__)", {
                    STORAGE: common.escapeAngularBrackets(storage.storageName),
                    DATACENTER: common.escapeAngularBrackets(storage.datacenterName)
                });
            })) : _("No storage is selected.");
    }

    function makeDatacentersOutput(profileDatacenters) {
        return !!profileDatacenters.length ? tags.ul(profileDatacenters.map(function (datacenter) {
                return datacenter.name;
            })) : _("No datacenter is selected.");
    }

    /**
     * Returns array of datacenters objects.
     * @param globals
     */
    function loadDatacenters(globals) {
        return xhr.get("/aps/2/resources/" + globals.aps.id + "/datacenters");
    }

    /**
     * Returns array of storages objects.
     * @param datacenters
     * @returns {*}
     */
    function loadStorages(datacenters) {
        var deferred = new Deferred(),
            result = [],
            requests = {};
        datacenters.forEach(function (datacenter) {
            requests[datacenter.aps.id] = xhr.get("/aps/2/resources/" + datacenter.aps.id + "/storages");
        });
        all(requests).then(function (storages) {
            datacenters.forEach(function (datacenter) {
                storages[datacenter.aps.id].forEach(function (storage) {
                    result.push({
                        id: datacenter.id + '_' + storage.id,
                        datacenterId: datacenter.id,
                        datacenterName: datacenter.name,
                        storageId: storage.id,
                        storageName: storage.name
                    });
                });
            });
            deferred.resolve(result);
        }, deferred.reject);
        return deferred;
    }

    /**
     * Returns of storages options list.
     * @returns {Array}
     * @param storages
     */
    function makeStoragesOptions(storages) {
        var options = [];
        storages.forEach(function (storage) {
            options.push({
                value: storage.id,
                label: _("__STORAGE__ (__DATACENTER__)", {
                    STORAGE: common.escapeAngularBrackets(storage.storageName),
                    DATACENTER: common.escapeAngularBrackets(storage.datacenterName)
                }),
                datacenterId: storage.datacenterId,
                datacenterName: storage.datacenterName,
                storageId: storage.storageId,
                storageName: storage.storageName
            });
        });
        return options;
    }

    /**
     * Returns of datacenters options list.
     * @returns {Array}
     * @param datacenters
     */
    function makeDatacentersOptions(datacenters) {
        var options = [];
        datacenters.forEach(function (datacenter) {
            options.push({
                value: datacenter.id,
                label: _("__DATACENTER__", {
                    DATACENTER: common.escapeAngularBrackets(datacenter.name)
                }),
                datacenterName: datacenter.name
            });
        });
        return options;
    }

    return {
        loadDatacenters: loadDatacenters,
        loadStorages: loadStorages,
        filterStorages: filterStorages,
        filterDatacenters: filterDatacenters,
        makeStoragesOutput: makeStoragesOutput,
        makeDatacentersOutput: makeDatacentersOutput,
        makeDatacentersOptions: makeDatacentersOptions,
        makeStoragesOptions: makeStoragesOptions
    };
});