/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

/**
 * Renders view user page controls.
 */
define([
    "dojo/Deferred",
    "aps/i18n",
    "aps/xhr",
    "app/common",
    "app/quotas"
], function (
    Deferred,
    _,
    xhr,
    common,
    quotas
) {
    function createQuotaOutput(counter, limit, usage) {
        var outputOptions = {
            label: quotas.formatStorageLabel(counter.label, limit)
        };

        switch (counter.legendSchema) {
            case 'feature':
                // Why always enabled ?
                outputOptions.value = _('Enabled');
                break;
            case 'storage':
                outputOptions.value = (usage ? Math.round(usage / common.BYTES_IN_GB) : '0') + ' / ' + (limit === common.QUOTA_UNLIMITED_VALUE ? _('Unlimited') : limit / common.BYTES_IN_GB);
                break;
            default:
                outputOptions.value = usage + ' / ' + (limit === common.QUOTA_UNLIMITED_VALUE ? _('Unlimited') : limit);
        }

        return ["aps/Output", outputOptions];
    }

    function printUserStatus(status) {
        switch (status) {
            case common.USER_STATUS.ACTIVE:
                return _('Enabled');
            case common.USER_STATUS.IN_PROGRESS:
                return _('In progress');
            case common.USER_STATUS.INACTIVE:
                return _('Disabled');
            case common.USER_STATUS.NOT_ACTIVATED:
                return _('Not activated');
            default:
                return _('Disabled');
        }
    }

    function printNotification(notifications, value) {
        return notifications.includes(value) ? _('Yes') : _('No');
    }

    function renderControls(account, user, showBack, userQuota, tenant, statistics) {
        var showQuotas = quotas.canShowQuotas(userQuota, account),
            countersList = common.getAvailableCountersForArea('quota_controls_ccpv1_view_user'),
            quotaOutputs = [];

        if (showQuotas) {
            for (let counter of countersList) {
                if (!(tenant[counter.name] && tenant[counter.name].limit)) {
                    continue;
                }

                var isSecondary = counter.name.indexOf('Secondary') > -1;
                var limit = isSecondary ? -1 : 0, usage = 0, counterStatistics = statistics[counter.name];
                if (counterStatistics) {
                    limit = counterStatistics.limit;
                    usage = counterStatistics.usage;
                }

                if (counter.legendSchema === 'storage') {
                    counter.label = quotas.formatStorageLabel(counter.label, limit);
                }
                if (counter.legendSchema === 'feature' || counterStatistics || isSecondary) {
                    quotaOutputs.push(createQuotaOutput(counter, limit, usage));
                }
            }
        }

        return ["aps/PageContainer", {id: "apsPageContainer"}, [
            ["aps/FieldSet", {
                title: _("Main")
            }, [
                ["aps/Output", {
                    label: _("User name"),
                    value: common.escapeAngularBrackets(account.displayName + ' (' + account.endUserLogin + ')')
                }],
                ["aps/Output", {
                    label: _('Administrator privileges'),
                    value: account.privileged ? _('Yes') : _('No')
                }]
            ]],
            ["aps/FieldSet", {
                escapeHTML: false,
                title: _('Account status')
            }, [
                ["aps/Output", {
                    label: _('Status'),
                    value: printUserStatus(account.userstatus)
                }]
            ]],
            ["aps/FieldSet", {
                escapeHTML: false,
                title: _('Notifications')
            }, [
                ["aps/FieldSet", {
                    label: _('MANAGEMENT PORTAL'),
                    title: false
                }],
                ["aps/Output", {
                    label: _('Quota overuse notifications'),
                    value: printNotification(account.notifications, 'quota')
                }],
                ["aps/Output", {
                    label: _('Scheduled usage reports'),
                    value: printNotification(account.notifications, 'reports')
                }],
                ["aps/Output", {
                    class: 'space-before-element'
                }],
                ["aps/FieldSet", {
                    label: _('CYBER PROTECTION'),
                    title: false
                }],
                ["aps/Output", {
                    label: _('Failure notifications'),
                    value: printNotification(account.notifications, 'backup_error')
                }],
                ["aps/Output", {
                    label: _('Warning notifications'),
                    value: printNotification(account.notifications, 'backup_warning')
                }],
                ["aps/Output", {
                    label: _('Success notifications'),
                    value: printNotification(account.notifications, 'backup_info')
                }],
                ["aps/Output", {
                    label: _('Daily recap about active alerts'),
                    value: printNotification(account.notifications, 'backup_daily_report')
                }]
            ]],
            ["aps/FieldSet", {
                escapeHTML: false,
                title: _('Quotas'),
                visible: showQuotas
            }, quotaOutputs],
            ["aps/FieldSet", {title: true}],
            ["aps/Container", {id: "buttons"}, [
                ["aps/Button", {
                    title: _('Edit'),
                    disabled: user.disabled,
                    onClick: function () {
                        var viewId = common.getCurrentView() === 'ccp--suservice--view' ?
                            'ccp--user--edit' :
                            common.getRelatedViewId('edit');
                        aps.apsc.gotoView(viewId, account.aps.id);
                    }
                }],
                ["aps/Button", {
                    title: _('Back to user list'),
                    visible: showBack,
                    onClick: function () {
                        aps.apsc.gotoView(common.getRelatedViewId('list'));
                    }
                }],
                ["aps/Button", {
                    title: _('Change Password'),
                    visible: account.userstatus == common.USER_STATUS.ACTIVE && !account.activateDuringCreation,
                    onClick: function () {
                        aps.apsc.gotoView(common.getRelatedViewId('change-password'), account.aps.id);
                    }
                }]
            ]]
        ]];
    }

    return {
        renderControls: renderControls
    };
});

