#!/usr/bin/python
import httplib
import json
import logging
import os
import sys
import time
import urlparse
import xmlrpclib
from optparse import OptionParser


class ConsoleHandler(logging.StreamHandler):
    """A handler that logs to console in the sensible way.

    StreamHandler can log to *one of* sys.stdout or sys.stderr.

    It is more sensible to log to sys.stdout by default with only error
    (logging.ERROR and above) messages going to sys.stderr. This is how
    ConsoleHandler behaves.
    """

    def __init__(self):
        logging.StreamHandler.__init__(self)
        self.stream = None  # reset it; we are not going to use it anyway

    def emit(self, record):
        if record.levelno >= logging.ERROR:
            self.__emit(record, sys.stderr)
        else:
            self.__emit(record, sys.stdout)

    def __emit(self, record, stream):
        self.stream = stream
        logging.StreamHandler.emit(self, record)
        self.stream = None

    def flush(self):
        # Workaround a bug in logging module
        # See:
        # http://bugs.python.org/issue6333
        if self.stream and hasattr(self.stream, 'flush') and not self.stream.closed:
            logging.StreamHandler.flush(self)


def check_rpc_response(response):
    if response['status'] != 0:
        raise Exception(response['error_message'])


def import_package(public_api, package_url):
    response = public_api.pem.APS.importPackage({"package_url": package_url})
    check_rpc_response(response)
    return response['result']['application_id']


def get_application_instances(public_api, application_id):
    response = public_api.pem.APS.getApplicationInstances({"app_id": application_id})
    check_rpc_response(response)
    return response['result']


def provide_application_instance(public_api, application_id, endpoint_url):
    response = public_api.pem.APS.provideApplicationInstance({
        "subscription_id": 0,
        "rt_id": 0,
        "app_id": application_id,
        "url_path": endpoint_url
    })
    check_rpc_response(response)
    application_instance_id = response['result']['application_instance_id']
    application_resource_id = response['result']['application_resource_id']
    return application_instance_id, application_resource_id


def upgrade_application_instance(public_api, application_instance_id, package_version):
    response = public_api.pem.APS.upgradeApplicationInstance({
        'application_instance_id': application_instance_id,
        'package_version': package_version
    })
    check_rpc_response(response)


def add_application_type(public_api, application_id, name, application_resource_id):
    response = public_api.pem.addResourceType({
        "name": name,
        "resclass_name": "rc.saas.service.link",
        "act_params": [
            {
                "var_name": "resource_uid",
                "var_value": application_resource_id
            },
            {
                "var_name": "app_id",
                "var_value": str(application_id)
            }
        ]
    })
    check_rpc_response(response)
    return response['result']['resource_type_id']


def add_service_type(public_api, application_id, name, service, autoprovide=False):
    response = public_api.pem.addResourceType({
        "name": name,
        "resclass_name": "rc.saas.service",
        "act_params": [
            {
                "var_name": "service_id",
                "var_value": service
            },
            {
                "var_name": "app_id",
                "var_value": str(application_id)
            },
            {
                "var_name": "autoprovide_service",
                "var_value": "1" if autoprovide else "0"
            }
        ]
    })
    check_rpc_response(response)
    return response['result']['resource_type_id']


def add_counter_type(public_api, application_id, name, service, counter, unit=False):
    response = public_api.pem.addResourceType({
        "name": name,
        "resclass_name": ("rc.saas.resource." + unit) if unit else "rc.saas.resource",
        "act_params": [
            {
                "var_name": "resource_id",
                "var_value": counter
            },
            {
                "var_name": "service_id",
                "var_value": service
            },
            {
                "var_name": "app_id",
                "var_value": str(application_id)
            }
        ]
    })
    check_rpc_response(response)
    return response['result']['resource_type_id']


def add_service_template(public_api, service_template_name, resources, category=None):
    types = []
    for k, v in resources.iteritems():
        types.append({'resource_type_id': k})

    parameters = {
        "owner_id": 1,
        "name": service_template_name,
        "resources": types
    }
    if category is not None:
        parameters["category"] = category

    response = public_api.pem.addServiceTemplate(parameters)
    check_rpc_response(response)
    service_template_id = response['result']['st_id']

    limits = []
    for k, v in resources.iteritems():
        limits.append({"resource_id": k, "resource_limit": v})

    response = public_api.pem.setSTRTLimits({"st_id": service_template_id, "limits": limits})
    check_rpc_response(response)

    response = public_api.pem.activateST({"st_id": service_template_id})
    check_rpc_response(response)

    return service_template_id

def get_resource_type_by_id(aps, id):
    if not hasattr(get_resource_type_by_id, "cache"):
        get_resource_type_by_id.cache = dict()

    if get_resource_type_by_id.cache.get(id) is None:
        aps.request("GET", "/aps/2/services/resource-type-manager/resourceTypes/{0}".format(id), '', {
            "APS-Token": aps.token
        })

        res = aps.getresponse()
        bp = json.loads(res.read())
        get_resource_type_by_id.cache[id] = bp['activationParameters']['resource_id']

    return get_resource_type_by_id.cache[id]

def resource_exists(public_api, name, resource_class_name, aliases=None, type=None,aps=None):
    response = public_api.pem.getResourceTypesByClass({
        "resclass_name": resource_class_name
    })

    check_rpc_response(response)
    names = [] if aliases is None else aliases
    names.append(name)
    for item in response['result']:
        if type is not None and aps is not None :
            resource_type = get_resource_type_by_id(aps,item['resource_type_id'])
            if type == resource_type:
                return True

        if item['resource_type_name'] in names:
            return True
    return False


def service_type_exists(public_api, name, aliases=None):
    return resource_exists(public_api, name, "rc.saas.service", aliases)


def counter_type_exists(public_api, aps, name, unit=False, aliases=None, type=None):
    return resource_exists(public_api, name, (("rc.saas.resource." + unit) if unit else "rc.saas.resource"), aliases, type, aps)


def install(public_api, application_id, application_resource_id, schema):
    application_type_id = add_application_type(public_api, application_id, schema['application'],
                                               application_resource_id)
    logging.info(
        "Resource type \"{0}\" has been created with ID {1}.".format(schema['application'], application_type_id))

    types = {}
    for service in schema['services']:
        resource_type_id = add_service_type(public_api, application_id, service['name'], service['type'],
                                            service['autoprovide'])
        logging.info("Resource type \"{0}\" has been created with ID {1}.".format(service['name'], resource_type_id))
        types[service['id']] = resource_type_id

        for counter in service['counters']:
            resource_type_id = add_counter_type(public_api, application_id, counter['name'], service['type'],
                                                counter['type'],
                                                counter['unit'])
            logging.info(
                "Resource type \"{0}\" has been created with ID {1}.".format(counter['name'], resource_type_id))
            types[counter['id']] = resource_type_id

    for profile in schema['profiles']:
        types[profile['id']] = add_profile(public_api, application_resource_id, profile['resource'])

    for template in schema['templates']:
        resources = {application_type_id: 1}
        for k, v in template['resources'].iteritems():
            resources[types[k]] = v
        service_template_id = add_service_template(public_api, template['name'], resources, template.get("category"))
        logging.info(
            "Service template \"{0}\" has been created with ID {1}.".format(template['name'], service_template_id))

    logging.info("Application \"{0}\" has been successfully installed.".format(schema['application']))


def upgrade_resources(application_id, public_api, schema, aps):
    for service in schema['services']:

        if not service_type_exists(public_api, service['name'], service.get('aliases')):
            resource_type_id = add_service_type(public_api, application_id, service['name'], service['type'],
                                                service['autoprovide'])
            logging.info(
                "Resource type \"{0}\" has been created with ID {1}.".format(service['name'], resource_type_id))
        for counter in service['counters']:
            if not counter_type_exists(public_api, aps, counter['name'], counter['unit'], counter.get('aliases'), counter['type']):
                resource_type_id = add_counter_type(public_api, application_id, counter['name'], service['type'],
                                                    counter['type'], counter['unit'])
                logging.info(
                    "Resource type \"{0}\" has been created with ID {1}.".format(counter['name'], resource_type_id))


def upgrade_instances(instances, package_version, public_api):
    for instance in instances:
        application_instance_id = instance['application_instance_id']
        logging.info("Upgrading instance with ID {0}...".format(application_instance_id))
        upgrade_application_instance(public_api, application_instance_id, package_version)
        logging.info("Instance with ID \"{0}\" has been successfully upgraded.".format(application_instance_id))


def load_schema(name):
    f = open(os.path.abspath(name), 'r')
    schema = json.load(f)
    return schema


def add_profile(public_api, application_resource_id, body):
    ret = public_api.pem.APS.getUserToken({"user_id": 1})
    check_rpc_response(ret)

    aps_token = ret['result']['aps_token']
    aps_uri = ret['result']['controller_uri']

    url = urlparse.urlparse(aps_uri)
    aps = httplib.HTTPSConnection(url.netloc)
    aps.connect()

    request_body = json.dumps(body)
    # foreach po resource
    aps.request("POST", "/aps/2/resources/{0}/storages".format(application_resource_id), request_body, {
        "APS-Token": aps_token
    })
    backup_profile_name = body['name']
    res = aps.getresponse()
    if res.status not in [200, 202]:
        logging.info("Can't create {0}".format(backup_profile_name))
        exit(1)

    bp = json.loads(res.read())
    aps.close()

    bp_id = bp['aps']['id']
    logging.info("'{0}' resource has been created with id: {1}".format(backup_profile_name, bp_id))

    bp_status = bp['aps']['status']
    while bp_status == 'aps:provisioning':
        time.sleep(1)
        aps.connect()
        aps.request("GET", "/aps/2/resources/{0}".format(bp_id), None, {"APS-Token": aps_token})
        res = aps.getresponse()
        if res.status != 200:
            logging.info("Can't create {0} resource type".format(backup_profile_name))
            exit(1)

        bp = json.loads(res.read())
        aps.close()
        bp_status = bp['aps']['status']

    profile_resource_id = bp['resourceTypeId']
    logging.info("'{0}' resource type has been created with id: {1}".format(backup_profile_name, profile_resource_id))
    return profile_resource_id

def get_aps_connection(public_api):
    ret = public_api.pem.APS.getUserToken({"user_id": 1})
    check_rpc_response(ret)

    aps_token = ret['result']['aps_token']
    aps_uri = ret['result']['controller_uri']

    url = urlparse.urlparse(aps_uri)
    aps = httplib.HTTPSConnection(url.netloc)
    aps.token = aps_token
    aps.connect()
    return aps

def main():
    parser = OptionParser()
    parser.add_option('-u', '--api-url', dest='api_url',
                      help='Public API access point, i.e., "https://localhost:8440/RPC2"')
    parser.add_option('-e', '--endpoint-url', dest='endpoint_url',
                      help='Application API end-point, i.e., "https://endpoint:11445/app"')
    parser.add_option('-p', '--package-name', dest='package_name',
                      help='Application package name, i.e., "application-1.0.0-xxx.app.zip"')
    parser.add_option('-l', '--log-file', dest='log_file', default="install.log",
                      help='Log file name, i.e., "/var/log/install.log"')
    parser.add_option('-r', '--resources', dest='resources_scheme', default="resources.json",
                      help='Path to resource scheme, i.e., "/var/resources.json"')
    parser.add_option('-v', '--version', dest='package_version',
                      help='Application package version, i.e., "1.0-6"')
    opts, args = parser.parse_args()

    if not (opts.api_url and opts.endpoint_url and opts.package_name):
        parser.print_help()
        exit(1)
        return

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    ch = ConsoleHandler()
    ch.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    logger.addHandler(ch)

    fh = logging.FileHandler(opts.log_file)
    fh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    logger.addHandler(fh)

    package_url = opts.endpoint_url + '/' + opts.package_name

    logging.info("Public API access point: {0}".format(opts.api_url))
    logging.info("Application API end-point: {0}".format(opts.endpoint_url))
    logging.info("Application package name: {0}".format(package_url))
    logging.info("Log file name: {0}".format(os.path.abspath(opts.log_file)))
    logging.info("Scheme file name: {0}".format(os.path.abspath(opts.resources_scheme)))
    logging.info("Application package version: {0}".format(opts.package_version))

    try:
        schema = load_schema(opts.resources_scheme)
        package_version = opts.package_version
        public_api = xmlrpclib.ServerProxy(opts.api_url)
        aps = get_aps_connection(public_api)
        logging.info("Importing APS application...")
        application_id = import_package(public_api, package_url)
        logging.info("Application \"{0}\" has been imported with ID {1}.".format(schema['application'], application_id))

        instances = get_application_instances(public_api, application_id)
        if len(instances):
            logging.info("Upgrading APS application...")
            upgrade_instances(instances, package_version, public_api)
            if schema["auto_install"]:
                upgrade_resources(application_id, public_api, schema, aps)
            logging.info("Application \"{0}\" has been successfully upgraded.".format(schema['application']))
            return

        logging.info("Installing APS application...")
        application_instance_id, application_resource_id = provide_application_instance(public_api, application_id,
                                                                                        opts.endpoint_url)

        logging.info("Application \"{0}\" has been instantiated with ID {1} ({2}).".format(schema['application'],
                                                                                           application_instance_id,
                                                                                           application_resource_id))

        if schema["auto_install"]:
            install(public_api, application_id, application_resource_id, schema)
    except Exception as e:
        logging.exception(e)
        exit(1)

if __name__ == '__main__':
    main()
