<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Base;

use Acronis\App\Base\ApplicationCounter as Counter;
use Acronis\App\CloudApi\Api;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Utils\CacheableTrait;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\App\Utils\UomConverter;
use Acronis\Cloud\Client\Model\Applications\Application;
use Acronis\Cloud\Client\Model\Infra\Infra;
use Acronis\Cloud\Client\Model\Infra\InfraCap;
use Acronis\Cloud\Client\Model\OfferingItems\OfferingItemOutput;
use Acronis\Cloud\Client\Model\OfferingItems\Quota;
use Acronis\Cloud\Client\Model\Tenants\OfferingItem;
use Acronis\Cloud\Client\Model\Usages\UsageOutput;
use Arr;
use Log;
use Model_Resource_TenantInterface as TenantStrategyInterface;
use Prophecy\Exception\InvalidArgumentException;
use Rest\RestException;
use Service_MspApiClient_ApiInterface as MspApiInterface;
use Str;

/**
 * todo: ADC APS 3.0. Convert this manager to class "Application" that extends \APS\BaseResource
 * Class ApplicationManager
 *
 * @package Acronis\App\CloudApi
 */
final class ApplicationsManager
{
    const INFRA_CAPABILITY_NOTARY = 'notary';

    use CacheableTrait,
        CountersManagerAwareTrait,
        DependenciesTrait,
        LegacyOfferingItemsTrait,
        MemoizeTrait;

    /** @var \tenant */
    protected $tenant;
    /** @var MspApiInterface */
    private $mspApi;
    /** @var string */
    private $tenantId;
    /** @var TenantStrategyInterface */
    private $tenantStrategy;

    private $skipOiValidation = false;

    const AVAILABLE_APPLICATION_TYPES = [
        'backup', 'notary', 'files_cloud', 'physical_data_shipping'
    ];

    private static $capabilitiesToApplicationTypes = [
        'backup' => 'backup',
        'disaster_recovery' => 'backup',
        'notary' => 'notary',
        'files_cloud' => 'files_cloud',
    ];

    /**
     * Application constructor.
     *
     * @param \tenant $tenant
     * @param MspApiInterface $mspApi
     * @param string|null $tenantId
     * @param TenantStrategyInterface $tenantStrategy
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function __construct(
        $tenant,
        MspApiInterface $mspApi,
        $tenantId = null,
        TenantStrategyInterface $tenantStrategy = null
    ) {
        $this->mspApi = $mspApi;
        $this->tenant = $tenant;
        $this->tenantId = $tenantId ?: $tenant->tenantId;
        $this->tenantStrategy = $tenantStrategy;
    }

    public function enableOfferingItemsValidationSkip()
    {
        $this->skipOiValidation = true;
    }

    public function getTenantEditions($applicationTypes = []): array
    {
        if (count($applicationTypes) === 0) {
            $applicationTypes = static::AVAILABLE_APPLICATION_TYPES;
        }
        $editions = [];
        foreach ($applicationTypes as $applicationType) {
            $applicationId = $this->getApplicationIdByType($applicationType);
            foreach ($this->getCloudApi()->getTenantApplicationOfferingItems($this->getTenantId(), $applicationId) as $oi) {
                if ($oi->getStatus() === ApiInterface::OFFERING_ITEM_STATUS_ACTIVE &&
                    $oi->getEdition() !== null &&
                    !in_array($oi->getEdition(), $editions)) {
                    $editions[] = $oi->getEdition();
                }
            }
        }
        return $editions;
    }


    /**
     * todo: ADC APS 3.0. use in method 'configure' for customers and partners
     *
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function updateOfferingItems()
    {
        $isApplicationEmpty = [];
        $usersOIs = [];
        foreach (static::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $targetEditions = [];
            $offeringItems = $this->buildOfferingItemsForApplication($applicationType, $targetEditions);
            $applicationId = $this->getApplicationIdByType($applicationType);
            if (!empty($offeringItems)) {
                $this->getCloudApi()->enableApplicationForTenantsChain($this->getRootTenantId(), $this->getTenantId(), $applicationId);
            }

            if ($this->tenant->type === \tenant::TYPE_RESELLER) {
                $this->enableApplicationOfferingItems($applicationId, $offeringItems);
            } else {
                $targetEdition = null;
                if ($applicationType === 'backup') {
                    foreach ($targetEditions as $edition) {
                        if (is_string($edition) && str_starts_with($edition, 'pck_per_')) {
                            $targetEdition = $edition;
                            break;
                        }
                    }
                }
                $usersOIs[$applicationId] = $this->updateApplicationOfferingItems($applicationType, $offeringItems, $targetEdition);
            }
            $isApplicationEmpty[$applicationId] = empty($offeringItems);
        }

        // having all applications empty still allows the creation of backup(by default), so we want backup enabled
        if (!in_array(false, $isApplicationEmpty)) {
            return $usersOIs;
        }

        foreach ($isApplicationEmpty as $applicationId => $isEmpty) {
            if ($isEmpty && $this->getCloudApi()->isApplicationEnabledForTenant($this->getTenantId(), $applicationId)) {
                $this->getCloudApi()->disableApplicationForTenant($this->getTenantId(), $applicationId);
            }
        }
        return $usersOIs;
    }

    /**
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function updateFolderOfferingItems()
    {
        $editions = [];
        foreach (static::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $offeringItems = $this->buildOfferingItemsForApplication($applicationType, $editions);
            $this->enableApplicationOfferingItems($this->getApplicationIdByType($applicationType), $offeringItems);
        }
    }

    public function fetchLimitsAndUsages()
    {
        $limits = $this->obtainCountersLimits();
        $usages = $this->obtainCountersUsages();

        return Arr::merge([
            'limit' => $limits,
            'usage' => $usages,
        ], [
            'limit' => 0,
            'usage' => 0,
        ]);
    }

    /**
     * @param string[] $excludedOfferingItems
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function deactivateAllTenantOfferingItems(array $excludedOfferingItems = [])
    {
        $tenantId = $this->getTenantId();
        $cloudApi = $this->getCloudApi();

        $tenantOfferingItems = [];
        foreach (static::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $tenantOfferingItems = array_merge(
                $tenantOfferingItems,
                $cloudApi->getTenantApplicationOfferingItems($tenantId, $this->getApplicationIdByType($applicationType))
            );
        }

        $excludedOfferingItemsHashtable = array_flip($excludedOfferingItems);

        $offeringItemsToUpdate = [];
        foreach ($tenantOfferingItems as $offeringItem) {
            if ($offeringItem->getStatus() === ApiInterface::OFFERING_ITEM_STATUS_INACTIVE) {
                continue;
            }

            $offeringItemName = $offeringItem->getName();
            if (isset($excludedOfferingItemsHashtable[$offeringItemName])) {
                continue;
            }

            $oi = new OfferingItem();
            $oi->setApplicationId($offeringItem->getApplicationId());
            $oi->setName($offeringItemName);
            $oi->setStatus(ApiInterface::OFFERING_ITEM_STATUS_INACTIVE);
            if ($offeringItem->hasInfraId()) {
                $oi->setInfraId($offeringItem->getInfraId());
            }

            $offeringItemsToUpdate[] = $oi;
        }

        $cloudApi->updateTenantApplicationOfferingItems($tenantId, $offeringItemsToUpdate);
    }

    /**
     * @param $type string Application type
     * @return string
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function getApplicationIdByType($type): string
    {
        return $this->fromCache(function() use ($type) {
            $application = $this->getCloudApi()->getApplicationByType($type);
            if (is_null($application)) {
                throw new RestException(404, Str::format(
                    'Cannot fetch application "%s".',
                    $type
                ));
            }
            return $application->getId();
        }, null, false, $type);
    }

    /**
     * @return TenantStrategyInterface
     */
    protected function getTenantStrategy()
    {
        return $this->tenantStrategy;
    }
    /**
     * @return Api
     */
    protected function getCloudApi()
    {
        return $this->memoize(function () {
            $mspApi = $this->getMspApi();

            return new Api(
                $mspApi->getUrl(),
                $mspApi->getLogin(),
                $mspApi->getPassword()
            );
        });
    }

    /**
     * @return MspApiInterface
     */
    protected function getMspApi()
    {
        return $this->mspApi;
    }

    /**
     * Returns active offering items only
     *
     * @param $applicationType
     * @param $editions
     * @return OfferingItem[]
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function buildOfferingItemsForApplication($applicationType, &$editions)
    {
        $applicationOfferingItems = [];
        $limits = $this->obtainCountersLimits($applicationType);
        foreach ($limits as $counterName => $limit) {
            // skip counters with limit 0
            if ($limit === 0) {
                continue;
            }

            $counter = $this->getCountersManager()->getCounterByName($counterName);

            if (!in_array($counter->getEdition(), $editions)) {
                $editions[] = $counter->getEdition();
            }

            $this->checkCounterDependencies($counter, $limits);

            $counterOfferingItems = $this->buildOfferingItemsForCounter($counter, $limit);

            $applicationOfferingItems = array_merge($applicationOfferingItems, $counterOfferingItems);
        }

        return $applicationOfferingItems;
    }

    /**
     * @return array
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function obtainCountersUsages()
    {
        if ($this->tenant->type === \tenant::TYPE_CUSTOMER) {
            return $this->obtainCountersUsagesFromCloud();
        }

        $usages = [];
        foreach (static::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $usages = array_merge($usages, $this->obtainCountersUsagesFromSubscription($applicationType));
        }

        return $usages;
    }

    /**
     * @param string $applicationType
     * @return array
     */
    protected function obtainCountersUsagesFromSubscription($applicationType)
    {
        $resources = $this->fetchSubscriptionResourcesForApplication($applicationType);
        $counters = $this->getApplicationCountersForUsages($applicationType);

        $countersHashtable = array_flip(array_map(function ($counter) {
            /** @var Counter $counter */
            return $counter->getName();
        }, $counters));

        $resources = array_filter($resources, function ($resource) use ($countersHashtable) {
            return Arr::has($countersHashtable, $resource->property);
        });

        return Arr::map($resources, 'property', function ($resource) {
            return $this->getUsageFromSubscriptionResource($resource);
        });
    }

    /**
     * @return array
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function obtainCountersUsagesFromCloud()
    {
        $applicationUsages = $this->fetchTenantAggregatedUsages();

        $counters = [];
        foreach (static::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $counters = array_merge($counters, $this->getApplicationCountersForUsages($applicationType));
        }

        $countersUsages = [];
        foreach ($counters as $counter) {
            $counterName = $counter->getName();
            $offeringItemName = $counter->getOfferingItemName();
            $offeringItemValue = Arr::get($applicationUsages, $offeringItemName, 0);

            $countersUsages[$counterName] = (int)round($this->getCountersManager()->convertValueFromOfferingItemUnits(
                $counter, $offeringItemValue
            ));
        }

        return $countersUsages;
    }

    /**
     * @param Counter $counter
     * @param int|float|null $limit
     * @return OfferingItem[]
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function buildOfferingItemsForCounter(Counter $counter, $limit)
    {
        if (!$this->checkOfferingItemIsEnabledForRoot($counter->getOfferingItemName())) {
            if ($this->tenant->type === \tenant::TYPE_RESELLER) {
                return [];
            }
            throw new RestException(403, Str::format(
                'Unable to provision resource "%s" if offering item "%s" is not available for tenant "%s".',
                $counter->getName(), $counter->getOfferingItemName(), $this->getRootTenantId()
            ));
        }

        $quotaValue = $counter->hasLimit()
            ? $this->getCountersManager()->convertValueToOfferingItemUnits($counter, $limit)
            : false;

        $offeringItems = [];
        if ($counter->isInfra()) {
            $offeringItemInfras = $this->resolveInfrasForCounter($counter);
            if (empty($offeringItemInfras)) {
                if ($this->tenant->type === \tenant::TYPE_RESELLER) {
                    return [];
                }
                throw new RestException(404, Str::format(
                    'No infrastructure components are available for offering item "%s".',
                    $counter->getOfferingItemName()
                ));
            }
            foreach ($offeringItemInfras as $infraId) {
                $offeringItems[] = $this->createOfferingItem($counter->getOfferingItemName(), $quotaValue, $infraId);
            }
        } else {
            $offeringItems[] = $this->createOfferingItem($counter->getOfferingItemName(), $quotaValue);
        }

        if ($this->tenantStrategy instanceof \Model_Resource_Tenant_Partner ||
            $this->tenantStrategy instanceof \Model_Resource_Tenant_Reseller
        ) {
            foreach ($counter->getPartnerOfferingItems() as $offeringItemName) {
                if (!$this->checkOfferingItemIsSupportedForTenantType($offeringItemName)) {
                    continue;
                }

                if (!$this->checkOfferingItemIsEnabledForRoot($offeringItemName)) {
                    continue;
                }

                $offeringItems[] = $this->createOfferingItem($offeringItemName);
            }
        }

        return $offeringItems;
    }

    /**
     * @param $applicationType string Application type
     * @return object[]
     */
    protected function fetchSubscriptionResourcesForApplication($applicationType)
    {
        return $this->memoize(function () use ($applicationType) {
            $resources = $this->tenant->subscription->resources();

            // get supported counters only
            $counters = $this->getApplicationCounters($applicationType);
            $countersHashtable = array_flip(array_map(function ($counter) {
                /** @var Counter $counter */
                return $counter->getName();
            }, $counters));

            // get counters only
            $apsType = dirname($this->tenant->aps->type);
            $resources = array_filter($resources, function ($resource) use ($apsType, $countersHashtable) {
                return isset($resource->apsType, $resource->property) &&
                    dirname($resource->apsType) === $apsType &&
                    Arr::has($countersHashtable, $resource->property);
            });

            //check that counters are unique
            $this->checkCountersAreUnique($resources);

            return $resources;
        }, $applicationType);
    }

    protected function getLimitFromSubscriptionResource($resource)
    {
        if (!isset($resource->limit)) {
            return Counter::UNLIMITED_VALUE;
        }

        $limit = (int)$resource->limit;
        if ($limit === 0) {
            return 0;
        }

        // Storage are always measured in Kb in subscription resources so convert them to bytes as defined in counters
        if (UomConverter::formatMeasurementName($resource->unit) === UomConverter::KILOBYTES) {
            $limit = UomConverter::convertStorage($limit, UomConverter::KILOBYTES, UomConverter::BYTES);
        }

        return $limit;
    }

    protected function getUsageFromSubscriptionResource($resource)
    {
        if (!isset($resource->usage)) {
            return 0;
        }

        $usage = (int)$resource->usage;
        if ($usage === 0) {
            return 0;
        }

        // Storage are always measured in Kb in subscription resources so convert them to bytes as defined in counters
        if (UomConverter::formatMeasurementName($resource->unit) === UomConverter::KILOBYTES) {
            $usage = UomConverter::convertStorage($usage, UomConverter::KILOBYTES, UomConverter::BYTES);
        }

        return $usage;
    }

    /**
     * @param string $offeringItemName
     * @param int|false $quotaValue
     * @param string|false $infraId
     * @return OfferingItem
     */
    protected function createOfferingItem($offeringItemName, $quotaValue = false, $infraId = false)
    {
        $offeringItem = new OfferingItem();
        $counter = $this->getCountersManager()->getCounterByOfferingItemName($offeringItemName);
        $offeringItem->setApplicationId($this->getApplicationIdByType($counter->getApplicationType()));
        $offeringItem->setName($offeringItemName);

        if ($infraId !== false) {
            $offeringItem->setInfraId($infraId);
        }

        if ($quotaValue === false) {
            return $offeringItem;
        }

        $quota = new Quota();
        $quota->setValue($quotaValue);
        $quota->setOverage(0);

        $offeringItem->setQuota($quota);

        return $offeringItem;
    }


    protected function getTenantId()
    {
        return $this->tenantId;
    }

    /**
     * @return string
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function getRootTenantId()
    {
        return $this->memoize(function () {
            $cloudApi = $this->getCloudApi();
            $me = $cloudApi->getMe();

            return $me->getTenantId();
        });
    }

    protected function getApplicationOfferingItemsNames($applicationType)
    {
        return $this->memoize(function () use ($applicationType) {
            $counters = $this->getApplicationCounters($applicationType);

            $offeringItemsNames = [];
            foreach ($counters as $counter) {
                $offeringItemsNames[] = $counter->getOfferingItemName();
                $offeringItemsNames = array_merge(
                    $offeringItemsNames,
                    $counter->getPartnerOfferingItems()
                );
            }

            return array_unique($offeringItemsNames);
        }, $applicationType);
    }

    /**
     * @param $applicationType string Application type
     * @return Counter[]
     */
    protected function getApplicationCounters($applicationType)
    {
        return $this->memoize(function () use ($applicationType) {
            $counters = $this->getCountersManager()->getApplicationCounters($applicationType);

            return array_filter($counters, function ($counter) {
                /** @var Counter $counter */
                $tenantTypes = $counter->getTenantTypes();

                return in_array($this->tenant->type, $tenantTypes);
            });
        }, $applicationType);
    }

    /**
     * @param $applicationType
     * @return Counter[]
     */
    protected function getApplicationCountersForUsages($applicationType)
    {
        return $this->memoize(function () use ($applicationType) {
            $counters = $this->getApplicationCounters($applicationType);
            $counters = array_filter($counters, function ($counter) {
                /** @var Counter $counter */
                return $counter->hasUsage();
            });

            return $counters;
        }, $applicationType);
    }

    /**
     * @param $applicationType
     * @return Counter[]
     */
    protected function getApplicationCountersForLimits($applicationType)
    {
        return $this->memoize(function () use ($applicationType) {
            $counters = $this->getApplicationCounters($applicationType);
            $counters = array_filter($counters, function ($counter) {
                /** @var Counter $counter */
                return $counter->isFeature()
                    || $counter->hasLimit();
            });

            return $counters;
        }, $applicationType);
    }

    protected function checkOfferingItemIsSupportedForTenantType($offeringItemName)
    {
        $this->tenant->checkType();
        $counter = $this->getCountersManager()->getCounterByOfferingItemName($offeringItemName);

        return $counter->isApplicableForTenantType($this->tenant->type);
    }

    /**
     * @param string $offeringItemName
     * @return boolean
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function checkOfferingItemIsEnabledForRoot($offeringItemName)
    {
        $applicationId = $this->getApplicationIdByOfferingItemName($offeringItemName);
        $offeringItems = $this->memoize(function () use ($applicationId) {
            $cloudApi = $this->getCloudApi();
            $me = $cloudApi->getMe();

            return $cloudApi->getTenantApplicationOfferingItems($me->getTenantId(), $applicationId);
        }, $applicationId);

        $offeringItems = array_filter($offeringItems, function ($offeringItem) use ($offeringItemName) {
            /** @var OfferingItem $offeringItem */
            return $offeringItem->getStatus() === Api::OFFERING_ITEM_STATUS_ACTIVE
                && $offeringItem->getName() === $offeringItemName;
        });

        return count($offeringItems) > 0;
    }

    /**
     * @param $tenantId
     * @param $applicationType
     * @param bool $force
     * @return \Acronis\Cloud\Client\Model\OfferingItems\OfferingItemOutput[]
     */
    protected function getTenantApplicationOfferingItems($tenantId, $applicationType, $force = false)
    {
        return $this->memoize(function () use ($tenantId, $applicationType) {
            $cloudApi = $this->getCloudApi();

            return $cloudApi->getTenantApplicationOfferingItems($tenantId, $this->getApplicationIdByType($applicationType));
        }, $tenantId . '-' . $applicationType, $force);
    }

    /**
     * @return UsageOutput[]
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function getTenantApplicationUsages()
    {
        return $this->memoize(function () {
            $usages = $this->getCloudApi()->getTenantUsages($this->getTenantId());

            return array_values(array_filter($usages, function ($usage) {
                /** @var UsageOutput $usage */
                return in_array($usage->getApplicationId(), array_values($this->getAvailableApplicationTypesToIds()));
            }));
        });
    }

    /**
     * @return mixed
     */
    private function getAvailableApplicationTypesToIds()
    {
        return $this->fromCache(function () {
            $applications = $this->getCloudApi()->getApplications();
            $supportedApplications = array_filter($applications, function($application) {
                /** @var Application $application */
                return in_array($application->getType(), static::AVAILABLE_APPLICATION_TYPES);
            });

            return Arr::map($supportedApplications, function($application) {
                /** @var Application $application */
                return $application->getType();
            }, function($application) {
                /** @var Application $application */
                return $application->getId();
            });
        });
    }

    /**
     * @param bool $absolute
     * @return array
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function fetchTenantAggregatedUsages($absolute = false)
    {
        $tenantUsages = $this->getTenantApplicationUsages();

        $aggregatedUsages = [];
        foreach ($tenantUsages as $offeringItemUsage) {
            // aggregate all usages by a offering item name
            $offeringItemName = $offeringItemUsage->getName();
            $offeringItemValue = $absolute
                ? $offeringItemUsage->getAbsoluteValue()
                : $offeringItemUsage->getValue();
            if (isset($aggregatedUsages[$offeringItemName])) {
                $aggregatedUsages[$offeringItemName] += $offeringItemValue;
            } else {
                $aggregatedUsages[$offeringItemName] = $offeringItemValue;
            }
        }

        return $aggregatedUsages;
    }

    /**
     * @param $applicationId
     * @param OfferingItem[] $offeringItems
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function enableApplicationOfferingItems($applicationId, array $offeringItems)
    {
        $cloudApi = $this->getCloudApi();
        $cloudApi->enableApplicationOfferingItemsForTenantsChain(
            $this->getRootTenantId(),
            $this->getTenantId(),
            $applicationId,
            $offeringItems
        );
    }

    /**
     * @param string $applicationType
     * @param OfferingItem[] $offeringItemsToEnable
     * @param string|null $targetEdition
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function updateApplicationOfferingItems($applicationType, array $offeringItemsToEnable, string $targetEdition = null)
    {
        // Disable other offering items
        $offeringItemsToUpdate = array_merge(
            $offeringItemsToEnable,
            $this->resolveOfferingItemsToDisable($applicationType, $offeringItemsToEnable)
        );

        $cloudApi = $this->getCloudApi();
        return $cloudApi->updateApplicationOfferingItemsForTenantsChain(
            $this->getRootTenantId(),
            $this->getTenantId(),
            $this->getApplicationIdByType($applicationType),
            $offeringItemsToUpdate,
            $targetEdition
        );
    }

    /**
     * @param $applicationType
     * @param OfferingItem[] $offeringItemsToEnable
     * @return OfferingItem[]
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function resolveOfferingItemsToDisable($applicationType, array $offeringItemsToEnable)
    {
        $offeringItemsToEnableHashtable = Api::getOfferingItemsHashTable($offeringItemsToEnable);
        $supportedOfferingItemsNamesHashTable = array_fill_keys(
            $this->getApplicationOfferingItemsNames($applicationType),
            true
        );

        $offeringItemNamesNormalizationMap  = $this->getCountersManager()->buildOfferingItemNameNormalizationMap();
        foreach ($offeringItemsToEnable as $offeringItem) {
            $offeringItemNameInStd = Arr::get($offeringItemNamesNormalizationMap, $offeringItem->getName());
            $enabledOfferingItemsNormalizedList[$offeringItemNameInStd] = true;
        }

        $offeringItemsToDisable = [];
        $tenantOfferingItems = $this->getTenantApplicationOfferingItems($this->tenantId, $applicationType);
        $tenantUsages = $this->getTenantApplicationUsages();
        $tenantUsagesHashmap = Arr::map($tenantUsages, function ($usageItem) {
            /** @var UsageOutput $usageItem */
            return $this->formatUsageKey($usageItem);
        }, function ($usageItem) {
            /** @var UsageOutput $usageItem */
            return $usageItem->getValue();
        });

        foreach ($tenantOfferingItems as $tenantOfferingItem) {
            $offeringItemHash = Api::getOfferingItemHash($tenantOfferingItem);
            if (Arr::has($offeringItemsToEnableHashtable, $offeringItemHash)) {
                continue;
            }

            // Never touch not supported offering items
            if (!Arr::has($supportedOfferingItemsNamesHashTable, $tenantOfferingItem->getName())) {
                continue;
            }

            // Don't touch already disabled offering items
            if ($tenantOfferingItem->getStatus() === ApiInterface::OFFERING_ITEM_STATUS_INACTIVE) {
                continue;
            }

            $offeringItem = $this->createOfferingItem(
                $tenantOfferingItem->getName(),
                $tenantOfferingItem->hasQuota() ? 0 : false,
                $tenantOfferingItem->hasInfraId() ? $tenantOfferingItem->getInfraId() : false
            );

            // Always disable features, non infra counters and infra counters if infra will have another enabled ois,
            // otherwise set zero quota
            if (Arr::get($tenantUsagesHashmap, $this->formatUsageKey($tenantOfferingItem), 0) != 0 &&
                !isset($enabledOfferingItemsNormalizedList[Arr::get($offeringItemNamesNormalizationMap, $tenantOfferingItem->getName())])) {

                if ($this->skipOiValidation) {
                    $offeringItem->setStatus(ApiInterface::OFFERING_ITEM_STATUS_INACTIVE);
                    $offeringItemsToDisable[] = $offeringItem;
                    Log::getInstance()->error('Found offering item "%s" which should be disabled in accordance with service plan but has not zero usage in Cloud.');
                    Log::getInstance()->error('Disabling offering item "%s" (possibly because of upgrade scenario).');

                    continue;
                }

                throw new RestException(409, Str::format(
                    'Can not disable offering item "%s" because it has not zero usage and it is not a switch to another edition for this offering item.',
                    $tenantOfferingItem->getName()
                ));
            } else {
                $offeringItem->setStatus(ApiInterface::OFFERING_ITEM_STATUS_INACTIVE);
                $offeringItemsToDisable[] = $offeringItem;
            }
        }

        return $offeringItemsToDisable;
    }

    /**
     * Method to get unique key for offering item usage for further check if it is 0 or not
     * @param UsageOutput|OfferingItemOutput $usageItem
     * @return string
     */
    private function formatUsageKey($usageItem) {
        // Name is already unique and contains edition prefix
        return $usageItem->getName() . '_' . ($usageItem->hasInfraId() ? $usageItem->getInfraId() : '');
    }

    /**
     * @param $offeringItemName
     * @return string
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function getApplicationIdByOfferingItemName($offeringItemName)
    {
        $offeringItemCounter = $this->getCountersManager()->getCounterByOfferingItemName($offeringItemName);

        if (null === $offeringItemCounter) {
            throw new RestException(404, Str::format(
                'Cannot fetch counter metadata for offering item "%s".',
                $offeringItemName
            ));
        }

        return $this->getApplicationIdByType($offeringItemCounter->getApplicationType());
    }

    /**
     * @param array $resources
     * @throws RestException
     */
    private function checkCountersAreUnique(array $resources)
    {
        $hashtable = [];
        foreach ($resources as $resource) {
            $counterName = $resource->property;
            if (isset($hashtable[$counterName])) {
                $conflictingResource = $hashtable[$counterName];
                throw new RestException(409, Str::format(
                    'Resources "%s" and "%s" have the same application counter "%s". Please remove one of them from the service template.',
                    $conflictingResource->id, $resource->id, $counterName
                )
                );
            }
            $hashtable[$counterName] = $resource;
        }
    }

    /**
     * @param null $applicationTypeToFilter
     * @return array
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function obtainCountersLimits($applicationTypeToFilter = null)
    {
        $limits = [];

        foreach (($applicationTypeToFilter === null ? static::AVAILABLE_APPLICATION_TYPES : [$applicationTypeToFilter]) as $applicationType) {
            $resources = $this->fetchSubscriptionResourcesForApplication($applicationType);
            $counters = $this->getApplicationCountersForLimits($applicationType);

            $countersHashtable = array_flip(array_map(function ($counter) {
                /** @var Counter $counter */
                return $counter->getName();
            }, $counters));

            $resources = array_filter($resources, function ($resource) use ($countersHashtable) {
                return Arr::has($countersHashtable, $resource->property);
            });

            $limits = array_merge($limits, Arr::map($resources, 'property', function ($resource) {
                return $this->getLimitFromSubscriptionResource($resource);
            }));
        }

        if ($this->tenant->type === \tenant::TYPE_RESELLER) {
            return $limits;
        }

        if ($applicationTypeToFilter === 'backup') {
            // business logic to keep the backward capability, can be removed when we sure that all partners upgraded to APS2.0
            $this->addMailboxInstanceSizeCounter($limits);
            $this->addLocalStorageUsageCounter($limits);
            $this->addO365MailboxesCounter($limits);
            $this->addO365OneDriveCounter($limits);
            $this->addO365SharePointSitesCounter($limits);
        }

        return $limits;
    }

    /**
     * @param ApplicationCounter $counter
     * @return string[]
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function resolveInfrasForCounter(Counter $counter)
    {
        $infraCapability = $counter->getInfraCapability();
        $edition = $counter->getEdition() ?: Api::EDITION_STD;

        if (is_null($infraCapability)) {
            return [];
        }

        return $this->getInfrasUuidsForCapability($infraCapability, $edition);
    }

    /**
     * @param string $infraCapability
     * @param null $edition
     * @return string[]
     */
    protected function getInfrasUuidsForCapability($infraCapability, $edition = null)
    {
        return $this->memoize(function () use ($infraCapability, $edition) {
            $offeringItemsNames = $this->getOfferingItemsNamesByCapability($infraCapability);
            $infrasUuids = $this->getTenantInfrasUuidsByOfferingItemsNames($offeringItemsNames, static::$capabilitiesToApplicationTypes[$infraCapability]);
            $editionInfras = [];
            if (is_null($edition)) {
                // since $infrasUuids is two dimensional array ([editionName => [infraIds]],
                // we flatten it to just have the values of the infra ids of all editions
                $editionInfras = array_unique(array_reduce($infrasUuids, function($allInfras, $editionInfras) {
                    return array_merge($allInfras, $editionInfras);
                }, []));
            } elseif (isset($infrasUuids[$edition])) {
                $editionInfras = $infrasUuids[$edition];
            }

            if (!empty($editionInfras)) {
                return $editionInfras;
            }

            return $this->resolveInfrasUuidsForCapability($infraCapability);
        }, $infraCapability . '-' . $edition);
    }

    /**
     * @param string $infraCapability
     * @return string[]
     */
    protected function getOfferingItemsNamesByCapability($infraCapability)
    {
        $applicationType = Arr::get(static::$capabilitiesToApplicationTypes, $infraCapability);
        if (null === $applicationType) {
            throw new \RuntimeException(sprintf('Unable to determine application type for infra capability "%s".', $infraCapability));
        }

        $counters = $this->getApplicationCountersForLimits($applicationType);
        $counters = array_values(array_filter(
            $counters,
            function ($counter) use ($infraCapability) {
                /** @var Counter $counter */
                return $counter->getInfraCapability() === $infraCapability;
            }
        ));

        return array_map(function ($counter) {
            /** @var Counter $counter */
            return $counter->getOfferingItemName();
        }, $counters);
    }

    /**
     * @param string[] $offeringItemsNames
     * @param $applicationType
     * @return string[]
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function getTenantInfrasUuidsByOfferingItemsNames(array $offeringItemsNames, $applicationType)
    {
        $tenantId = $this->tenant->type === \tenant::TYPE_CUSTOMER ? $this->getTenantId() : $this->getRootTenantId();
        $offeringItems = $this->getTenantApplicationOfferingItems($tenantId, $applicationType);
        $offeringItemsNamesHashtable = array_fill_keys($offeringItemsNames, true);
        $offeringItems = array_values(array_filter(
            $offeringItems,
            function ($offeringItem) use ($offeringItemsNamesHashtable) {
                /** @var OfferingItemOutput $offeringItem */
                return isset($offeringItemsNamesHashtable[$offeringItem->getName()])
                    && $offeringItem->getStatus() === Api::OFFERING_ITEM_STATUS_ACTIVE
                    && $offeringItem->hasInfraId();
            }
        ));

        $offeringItemInfras = [];
        /** @var OfferingItemOutput $offeringItem */
        foreach ($offeringItems as $offeringItem) {
            $edition = $offeringItem->getEdition();
            $infraId = $offeringItem->getInfraId();
            $offeringItemInfras[$edition] = Arr::get($offeringItemInfras, $edition, []);
            if (!in_array($infraId, $offeringItemInfras[$edition])) {
                $offeringItemInfras[$edition][] = $infraId;
            }
        }

        return $offeringItemInfras;
    }

    /**
     * @param $infraCapability
     * @return string[]
     * @throws \Rest\RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function resolveInfrasUuidsForCapability($infraCapability)
    {
        if ($infraCapability === InfraCap::BACKUP) {
            return $this->resolveBackupStorageInfrasUuids();
        }

        if ($infraCapability === InfraCap::DISASTER_RECOVERY) {
            return $this->resolveDisasterRecoveryInfrasUuids();
        }

        if (in_array($infraCapability, [InfraCap::FILES_CLOUD, static::INFRA_CAPABILITY_NOTARY])) {
            return $this->resolveInfrasUuidsAsFilteringByCapability($infraCapability);
        }

        throw new InvalidArgumentException(Str::format(
            'Unsupported infrastructure component capability "%s".',
            $infraCapability
        ));
    }

    /**
     * @param string $infraCapability
     * @return array
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function resolveInfrasUuidsAsFilteringByCapability($infraCapability)
    {
        $availableInfras = $this->getCloudApi()->fetchInfrasForTenantsChain($this->getRootTenantId(), $this->getTenantId());
        $suitableInfras = array_filter($availableInfras, function($infra) use ($infraCapability) {
            /** @var Infra $infra */
            return in_array($infraCapability, $infra->getCapabilities());
        });

        return array_values(array_map(function($infra) {
            /** @var Infra $infra */
            return $infra->getId();
        }, $suitableInfras));
    }

    /**
     * @return string[]
     * @throws \Rest\RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function resolveBackupStorageInfrasUuids()
    {
        $rootTenantId = $this->getRootTenantId();
        $leafTenantId = $this->getTenantId();

        $storages = $this->fetchBackupStorageUuidsForTenantsChain($rootTenantId, $leafTenantId);
        if ($this->tenant->type !== \tenant::TYPE_CUSTOMER) {
            return array_values($storages);
        }

        $tenantStrategy = $this->getTenantStrategy();
        $storageId = $tenantStrategy->resolveBackupStorageId();

        $storageUuid = is_numeric($storageId)
            ? Arr::get($storages, $storageId)
            : (in_array($storageId, array_values($storages)) ? $storageId : null);

        if (is_null($storageUuid)) {
            throw new \Rest\RestException(404, Str::format(
                'There is no backup storage with ID %s.',
                $storageId
            ));
        }

        return [$storageUuid];
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return string[]
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function fetchBackupStorageUuidsForTenantsChain($rootTenantId, $leafTenantId)
    {
        $cloudApi = $this->getCloudApi();
        $tenantPath = $cloudApi->resolveTenantPath($rootTenantId, $leafTenantId);

        $mspApi = $this->getMspApi();
        $storages = [];
        foreach ($tenantPath as $tenantId) {
            $storages = array_merge($storages, $mspApi->getStorages($tenantId));
        }

        return Arr::map(
            $storages,
            function ($storage) {
                /** @var \Service_MspApiClient_Response_Struct_Storage $storage */
                return $storage->getId();
            },
            function ($storage) {
                /** @var \Service_MspApiClient_Response_Struct_Storage $storage */
                return $storage->getUuid();
            },
            false
        );
    }

    /**
     * @return string[]
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function resolveDisasterRecoveryInfrasUuids()
    {
        $rootTenantId = $this->getRootTenantId();
        $leafTenantId = $this->getTenantId();
        $cloudApi = $this->getCloudApi();
        $availableInfras = $cloudApi->fetchInfrasForTenantsChain($rootTenantId, $leafTenantId);

        if ($this->tenant->type === \tenant::TYPE_RESELLER) {
            $disasterRecoveryInfras = array_filter($availableInfras, function($infra) {
                /** @var Infra $infra */
                return in_array(InfraCap::DISASTER_RECOVERY, $infra->getCapabilities());
            });
            return array_map(function($infra) {
                /** @var Infra $infra */
                return $infra->getId();
            }, $disasterRecoveryInfras);
        }

        $backupStorageInfrasUuids = $this->getInfrasUuidsForCapability(InfraCap::BACKUP);
        $backupStorageInfras = array_filter($availableInfras, function ($infra) use ($backupStorageInfrasUuids) {
            /** @var Infra $infra */
            return in_array($infra->getId(), $backupStorageInfrasUuids);
        });
        if (empty($backupStorageInfras)) {
            return [];
        }

        $locationsUuids = array_map(function ($infra) {
            /** @var Infra $infra */
            return $infra->getLocationId();
        }, $backupStorageInfras);

        $disasterRecoveryInfras = array_filter($availableInfras, function ($infra) use ($locationsUuids) {
            /** @var Infra $infra */
            return in_array($infra->getLocationId(), $locationsUuids)
                && in_array(InfraCap::DISASTER_RECOVERY, $infra->getCapabilities());
        });
        if (empty($disasterRecoveryInfras)) {
            return [];
        }

        $disasterRecoveryInfrasUuids = array_map(function ($infra) {
            /** @var Infra $infra */
            return $infra->getId();
        }, $disasterRecoveryInfras);

        return array_values($disasterRecoveryInfrasUuids);
    }
}