<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\CloudApi;

use Acronis\App\Base\ApplicationsManager;
use Acronis\App\Base\GroupIdConverterTrait;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\Cloud\Client\ApiException;
use Acronis\Cloud\Client\HttpException;
use Acronis\Cloud\Client\IOException;
use Acronis\Cloud\Client\Model\AccessPolicies\AccessPoliciesList;
use Acronis\Cloud\Client\Model\AccessPolicies\AccessPolicy;
use Acronis\Cloud\Client\Model\Applications\Application;
use Acronis\Cloud\Client\Model\Common\Contact\Contact;
use Acronis\Cloud\Client\Model\Infra\Infra;
use Acronis\Cloud\Client\Model\OfferingItems\OfferingItemOutput;
use Acronis\Cloud\Client\Model\OfferingItems\Quota;
use Acronis\Cloud\Client\Model\Pricing\TenantPricingSettings;
use Acronis\Cloud\Client\Model\Pricing\TenantPricingSettingsPut;
use Acronis\Cloud\Client\Model\Reports\Report;
use Acronis\Cloud\Client\Model\Reports\ReportPost;
use Acronis\Cloud\Client\Model\Reports\ReportPostKind;
use Acronis\Cloud\Client\Model\Reports\ReportPostLevel;
use Acronis\Cloud\Client\Model\Reports\ReportPostParameters;
use Acronis\Cloud\Client\Model\Reports\ReportPostResultAction;
use Acronis\Cloud\Client\Model\Reports\ReportPostSchedule;
use Acronis\Cloud\Client\Model\Reports\ReportPostType;
use Acronis\Cloud\Client\Model\Reports\Stored\StoredReportParamsItems;
use Acronis\Cloud\Client\Model\Tenants\OfferingItem;
use Acronis\Cloud\Client\Model\Tenants\OfferingItemsPut;
use Acronis\Cloud\Client\Model\Tenants\Tenant;
use Acronis\Cloud\Client\Model\Tenants\TenantPost;
use Acronis\Cloud\Client\Model\Tenants\TenantPostUpdateLock;
use Acronis\Cloud\Client\Model\Tenants\TenantPut;
use Acronis\Cloud\Client\Model\Tenants\TenantPutUpdateLock;
use Acronis\Cloud\Client\Model\Usages\UsageOutput;
use Acronis\Cloud\Client\Model\Users\User;
use Acronis\Cloud\Client\Model\Users\UserPost;
use Acronis\Cloud\Client\Model\Users\UserPut;
use Arr;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\GuzzleException;
use Log;
use Rest\RestException;
use Service_MspApiClient_Request_Struct_CreateGroupInterface as GroupInterface;
use Service_MspApiClient_Request_Struct_GroupPricing as GroupPricing;
use Str;

class Api extends AuthorizedApi implements ApiInterface
{
    use GroupIdConverterTrait;
    use MemoizeTrait;
    use \Service_MspApiClient_MspApiFactoryTrait;

    const CACHE_TTL = 7200;
    const URL_PATH_API_V1 = '/api/1';
    const URL_PATH_API_V2 = '/api/2';

    /** @var string */
    private $url;

    /** @var string */
    private $serverUrl;

    /** @var string */
    private $login;

    /** @var string */
    private $password;

    public function __construct($url, $login, $password)
    {
        $this->url = $url;
        $this->login = $login;
        $this->password = $password;
    }

    /**
     * @param UsageOutput|OfferingItem|OfferingItemOutput $offeringItem
     * @return string
     */
    public static function getOfferingItemHash($offeringItem)
    {
        return Str::format(
            '%s:%s',
            $offeringItem->getName(), $offeringItem->hasInfraId() ? $offeringItem->getInfraId() : ''
        );
    }

    /**
     * @param UsageOutput[]|OfferingItem[]|OfferingItemOutput[] $offeringItems
     * @return array
     */
    public static function getOfferingItemsHashTable(array $offeringItems)
    {
        return Arr::map(
            $offeringItems,
            function ($offeringItem) {
                return static::getOfferingItemHash($offeringItem);
            },
            function ($offeringItem) {
                return $offeringItem;
            }
        );
    }

    /**
     * @return string
     */
    public function getUrl()
    {
        return $this->url;
    }

    /**
     * @return string
     */
    public function getLogin()
    {
        return $this->login;
    }

    /**
     * @return string
     */
    public function getPassword()
    {
        return $this->password;
    }

    /* ====== Users API ====== */

    /**
     * @return User
     * @throws ApiException
     * @throws HttpException
     */
    public function getMe()
    {
        return $this->memoize(function () {
            return $this->authorizedCall(function () {
                return $this->getUsersApi()->getUsersMe();
            });
        });
    }

    /**
     * Returns True if the login exists otherwise False
     *
     * @param string $login
     * @return bool
     * @throws ApiException
     * @throws HttpException
     */
    public function checkLogin($login)
    {
        return $this->authorizedCall(function () use ($login) {
            try {
                $this->getUsersApi()->getUsersCheckLogin($login);

                return false;
            } catch (HttpException $e) {
                if ($e->getCode() != 409) {
                    throw $e;
                }

                return true;
            }
        });
    }

    /* ====== Methods to manage users ====== */

    /**
     * @param UserPost $userBody
     * @return User
     * @throws ApiException
     * @throws HttpException
     */
    public function createUser(UserPost $userBody)
    {
        return $this->authorizedCall(function () use ($userBody) {
            return $this->getUsersApi()->postUsers($userBody);
        });
    }

    /**
     * @param string $userId
     * @param int $version
     * @return User
     * @throws ApiException
     * @throws HttpException
     */
    public function deleteUser($userId, $version)
    {
        return $this->authorizedCall(function () use ($userId, $version) {
            return $this->getUsersApi()->deleteUsersByUserId($userId, $version);
        });
    }

    /**
     * @param string $userId
     * @return User
     * @throws ApiException
     * @throws HttpException
     */
    public function getUser($userId)
    {
        return $this->authorizedCall(function () use ($userId) {
            return $this->getUsersApi()->getUsersByUserId($userId);
        });
    }

    /**
     * @param string $userId
     * @param UserPut $userBody
     * @return User
     * @throws ApiException
     * @throws HttpException
     */
    public function updateUser($userId, UserPut $userBody)
    {
        return $this->authorizedCall(function () use ($userBody, $userId) {
            return $this->getUsersApi()->putUsersByUserId($userId, $userBody);
        });
    }

    /**
     * @param $userId
     * @param AccessPolicy[] $accessPolicies
     * @return AccessPolicy[]
     * @throws ApiException
     * @throws HttpException
     */
    public function updateUserAccessPolicies($userId, array $accessPolicies)
    {
        $body = new AccessPoliciesList();
        $body->setItems($accessPolicies);

        /**
         * @var AccessPoliciesList $response
         */
        $response = $this->authorizedCall(function () use ($userId, $body) {
            return $this->getUsersApi()->putUsersAccessPoliciesByUserId($userId, $body);
        });

        return $response->getItems();
    }

    /* ====== Methods to manage tenants ====== */

    const TENANT_KIND_MAPPING = [
        GroupInterface::EUC_GROUP_KIND_ID => 'customer',
        GroupInterface::RESELLER_GROUP_KIND_ID => 'partner',
        GroupInterface::FOLDER_GROUP_KIND_ID => 'folder',
    ];

    const PRICING_MAPPINGS = [
        GroupPricing::PRODUCTION_MODE => 'production',
        GroupPricing::TRIAL_MODE => 'trial',
    ];


    /**
     * @param $groupId int
     * @param $version int
     * @return HttpResponse
     * @throws \Exception
     */
    public function deleteTenant($tenantId, $version = null)
    {
        return $this->authorizedCall(function () use ($tenantId, $version) {
            return $this->getTenantsApi()->deleteTenantsByTenantId($tenantId, $version);
        });
    }


    /**
     * @param GroupInterface $groupInfo
     * @param $parentGroupId
     * @return \Acronis\Cloud\Client\Model\Tenants\Tenant|void
     * @throws \Exception
     */
    public function createTenant($groupInfo, $parentGroupId)
    {
        $tenant = new TenantPost();
        if (!array_key_exists($groupInfo->getKind(), static::TENANT_KIND_MAPPING)) {
            throw new \Exception(sprintf('Not supported tenant kind "%s" provided.',
                $groupInfo->getKind()));
        }
        if (!array_key_exists($groupInfo->getPricing()->getMode(), static::PRICING_MAPPINGS)) {
            throw new \Exception(sprintf('Not supported pricing mode "%s" provided.',
                $groupInfo->getPricing()->getMode()));
        }

        $groupKind = static::TENANT_KIND_MAPPING[$groupInfo->getKind()];
        $tenant->setKind($groupKind);
        $tenant->setName($groupInfo->getName());
        $contact = new Contact();
        $groupInfoContact = $groupInfo->getContact();
        $contact->setEmail($groupInfoContact->getEmail());
        $contact->setAddress1($groupInfoContact->getAddress1());
        $contact->setAddress2($groupInfoContact->getAddress2());
        $contact->setCity($groupInfoContact->getCity());
        $contact->setState($groupInfoContact->getState());
        $contact->setCountry($groupInfoContact->getCountry());
        $contact->setZipcode($groupInfoContact->getZipCode());
        $contact->setPhone($groupInfoContact->getPhone());
        $contact->setFirstname($groupInfoContact->getFirstName());
        $contact->setLastname($groupInfoContact->getLastName());
        $tenant->setContact($contact);
        $tenant->setParentId($parentGroupId);
        if ($groupKind !== static::TENANT_KIND_MAPPING[GroupInterface::FOLDER_GROUP_KIND_ID]) {
            $tenant->setCustomerId($groupInfo->getCustomerId());
        }
        if ($groupKind === 'customer') {
            $lock = new TenantPostUpdateLock();
            $lock->setEnabled(true);
            $tenant->setUpdateLock($lock);
        }

        /** @var \Acronis\Cloud\Client\Model\Tenants\Tenant $createdTenant */
        $createdTenant = $this->authorizedCall(function () use ($tenant) {
            return $this->getTenantsApi()->postTenants($tenant);
        });

        $tenantId = $createdTenant->getId();
        /** @var TenantPricingSettings $pricing */
        $currentPricing = $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsPricingByTenantId($tenantId);
        });
        $pricing = new TenantPricingSettingsPut();
        $pricing->setMode(static::PRICING_MAPPINGS[$groupInfo->getPricing()->getMode()]);
        $pricing->setVersion($currentPricing->getVersion());

        $this->authorizedCall(function () use ($pricing, $tenantId) {
            $this->getTenantsApi()->putTenantsPricingByTenantId($tenantId, $pricing);
        });

        return $createdTenant;
    }

    /**
     * @param string $tenantId
     * @return Tenant
     * @throws HttpException
     * @throws ApiException
     */
    public function getTenant($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsByTenantId($tenantId);
        });
    }


    /**
     * @param string $tenantId
     * @int $version
     * @return Tenant
     */

    public function enableTenant($tenantId, $version)
    {
        $tenantPut = new TenantPut();
        $tenantPut->setVersion($version);
        $tenantPut->setEnabled(true);
        return $this->updateTenant($tenantId, $tenantPut);
    }


    /**
     * @param string $tenantId
     * @int $version
     * @return Tenant
     */

    public function disableTenant($tenantId, $version)
    {
        $tenantPut = new TenantPut();
        $tenantPut->setVersion($version);
        $tenantPut->setEnabled(false);
        return $this->updateTenant($tenantId, $tenantPut);
    }

    /**
     * @param \Service_MspApiClient_Request_Struct_UpdateGroupInterface $groupInfo
     * @param string $tenantId
     * @param integer $version
     * @param $groupKind
     * @param $mspApi
     * @return integer
     * @throws ApiException
     * @throws HttpException
     */
    public function updateTenantFromGroupInfo($groupInfo, $tenantId, $version, $groupKind, $mspApi)
    {
        if (!array_key_exists($groupInfo->getPricing()->getMode(), static::PRICING_MAPPINGS)) {
            throw new \Exception(sprintf('Not supported pricing mode "%s" provided.',
                $groupInfo->getPricing()->getMode()));
        }

        $tenant = new TenantPut();
        $tenant->setVersion($version);
        (null !== $groupInfo->getParentId()) && $tenant->setParentId($this->convertToTenantId($groupInfo->getParentId(), $mspApi));
        (null !== $groupInfo->getName()) && $tenant->setName($groupInfo->getName());
        $contact = new Contact();
        $groupInfoContact = $groupInfo->getContact();
        (null !== $groupInfoContact->getEmail()) && $contact->setEmail($groupInfoContact->getEmail());
        (null !== $groupInfoContact->getAddress1()) && $contact->setAddress1($groupInfoContact->getAddress1());
        (null !== $groupInfoContact->getAddress2()) && $contact->setAddress2($groupInfoContact->getAddress2());
        (null !== $groupInfoContact->getCity()) && $contact->setCity($groupInfoContact->getCity());
        (null !== $groupInfoContact->getState()) && $contact->setState($groupInfoContact->getState());
        (null !== $groupInfoContact->getCountry()) && $contact->setCountry($groupInfoContact->getCountry());
        (null !== $groupInfoContact->getZipCode()) && $contact->setZipcode($groupInfoContact->getZipCode());
        (null !== $groupInfoContact->getPhone()) && $contact->setPhone($groupInfoContact->getPhone());
        (null !== $groupInfoContact->getFirstName()) && $contact->setFirstname($groupInfoContact->getFirstName());
        (null !== $groupInfoContact->getLastName()) && $contact->setLastname($groupInfoContact->getLastName());
        $tenant->setContact($contact);
        (null !== $groupInfo->getCustomerId()) && $tenant->setCustomerId($groupInfo->getCustomerId());
        if ($groupKind === GroupInterface::EUC_GROUP_KIND_ID) {
            $lock = new TenantPutUpdateLock();
            $lock->setEnabled(true);
            $tenant->setUpdateLock($lock);
        }

        /** @var \Acronis\Cloud\Client\Model\Tenants\Tenant $updatedTenant */
        $updatedTenant = $this->authorizedCall(function () use ($tenant, $tenantId) {
            return $this->getTenantsApi()->putTenantsByTenantId($tenantId, $tenant);
        });

        if ($groupInfo->getPricing()->getMode() === GroupPricing::PRODUCTION_MODE) {
            /** @var TenantPricingSettings $pricing */
            $currentPricing = $this->authorizedCall(function () use ($tenantId) {
                return $this->getTenantsApi()->getTenantsPricingByTenantId($tenantId);
            });
            $pricing = new TenantPricingSettingsPut();
            $pricing->setMode('production');
            $pricing->setVersion($currentPricing->getVersion());
            $this->authorizedCall(function () use ($pricing, $tenantId) {
                $this->getTenantsApi()->putTenantsPricingByTenantId($tenantId, $pricing);
            });
        }

        return $updatedTenant->getVersion();
    }

    /**
     * @param $tenantId
     * @param $putTenant
     * @return Tenant
     * @throws ApiException
     * @throws HttpException
     * todo we've faced a lof of errors while calling API with incorrect params (not only in case of APIv1/v2 tenant id format)
     * need to implement getting minified call stack in CloudApi.
     * Better to define some interface and setters to inject some logic of validation error mechanism in API, use default one
     * when not set - (in generated API lib)
     */
    public function updateTenant($tenantId, $putTenant)
    {
        return $this->authorizedCall(function () use ($tenantId, $putTenant) {
            return $this->getTenantsApi()->putTenantsByTenantId($tenantId, $putTenant);
        });
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function resolveTenantPath($rootTenantId, $leafTenantId)
    {
        $tenantId = $leafTenantId;
        $tenantPath = [$tenantId];
        while ($tenantId !== $rootTenantId) {
            $tenantData = $this->getTenant($tenantId);
            $tenantId = $tenantData->getParentId();

            $tenantPath[] = $tenantId;
        }

        return $tenantPath;
    }

    /**
     * @param string $tenantId
     * @return UsageOutput[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantUsages($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsUsagesByTenantId($tenantId);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @return User[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantUsers($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsUsersByTenantId($tenantId);
        })->getItems();
    }

    /* ====== Methods to manage tenant's applications ====== */

    /**
     * @return Application[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getApplications()
    {
        return $this->fromCache(function () {
            return $this->authorizedCall(function () {
                return $this->getApplicationsApi()->getApplications();
            })->getItems();
        });
    }

    /**
     * @param string $type
     * @return Application|null
     * @throws ApiException
     * @throws HttpException
     */
    public function getApplicationByType($type)
    {
        $applications = array_filter($this->getApplications(), function ($application) use ($type) {
            /** @var Application $application */
            return $application->getType() === $type;
        });

        return reset($applications) ?: null;
    }

    /**
     * @param string $tenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantApplicationsUuids($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsApplicationsByTenantId($tenantId);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @return bool
     * @throws ApiException
     * @throws HttpException
     */
    public function isApplicationEnabledForTenant($tenantId, $applicationId)
    {
        return in_array($applicationId, $this->getTenantApplicationsUuids($tenantId));
    }

    /**
     * @param $tenantId
     * @param $kind
     * @param string $edition
     * @return mixed
     * @throws ApiException
     * @throws HttpException
     */
    public function getOfferingItemsAvailableForChild($tenantId, $kind, $edition = self::EDITION_ANY)
    {
        return $this->authorizedCall(function () use ($tenantId, $kind, $edition) {
            return $this->getTenantsApi()->getTenantsOfferingItemsAvailableForChild($tenantId, $kind, $edition);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @throws ApiException
     * @throws HttpException
     */
    public function enableApplicationForTenant($tenantId, $applicationId)
    {
        $this->authorizedCall(function () use ($applicationId, $tenantId) {
            return $this->getApplicationsApi()->postApplicationsBindingsTenants($applicationId, $tenantId);
        });
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @throws ApiException
     * @throws HttpException
     */
    public function disableApplicationForTenant($tenantId, $applicationId)
    {
        $this->authorizedCall(function () use ($applicationId, $tenantId) {
            return $this->getApplicationsApi()->deleteApplicationsBindingsTenants($applicationId, $tenantId);
        });
    }

    /* ====== Methods to manage offering items ====== */

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @param string $edition
     * @return OfferingItemOutput[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantApplicationOfferingItems($tenantId, $applicationId, $edition = '*')
    {
        return $this->authorizedCall(function () use ($tenantId, $applicationId, $edition) {
            return $this->getTenantsApi()->getTenantsApplicationsOfferingItems($tenantId, $applicationId, $edition);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @param OfferingItem[] $offeringItems
     * @return OfferingItemOutput[]
     * @throws ApiException
     * @throws HttpException
     */
    public function updateTenantApplicationOfferingItems($tenantId, $offeringItems)
    {
        $body = new OfferingItemsPut();
        $body->setOfferingItems($offeringItems);

        return $this->authorizedCall(function () use ($tenantId, $body) {
            return $this->getTenantsApi()->putTenantsOfferingItemsByTenantId($tenantId, $body);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @param string $targetEdition
     * @return OfferingItemOutput[]
     * @throws ApiException
     * @throws HttpException
     */
    public function switchTenantEdition(string $tenantId, string $applicationId, string $targetEdition): array
    {
        $body = new \Acronis\Cloud\Client\Model\Editions\EditionSwitchPut();
        $body->setTargetEdition($targetEdition);
        $body->setApplicationId($applicationId);
        return $this->authorizedCall(function () use ($tenantId, $body) {
            return $this->getTenantsApi()->putTenantEditionSwitch($tenantId, $body);
        })->getItems();
    }

    /**
     * Recursive enable an application for a hierarchy of tenants except root tenant.
     *
     * @param string $rootTenantId
     * @param string $leafTenantId
     * @param string $applicationId
     * @throws ApiException
     * @throws HttpException
     */
    public function enableApplicationForTenantsChain($rootTenantId, $leafTenantId, $applicationId)
    {
        if ($leafTenantId === $rootTenantId) {
            return;
        }

        if ($this->isApplicationEnabledForTenant($leafTenantId, $applicationId)) {
            return;
        }

        // enable for the parent
        $tenant = $this->getTenant($leafTenantId);
        $this->enableApplicationForTenantsChain($rootTenantId, $tenant->getParentId(), $applicationId);
        // enable for the tenant
        $this->enableApplicationForTenant($leafTenantId, $applicationId);
    }

    /**
     * Recursive enable specified offering items with unlimited quotas for a hierarchy of tenants except root tenant.
     * Note: Enabled or not specified offering items won't be affected
     *
     * @param string $rootTenantId
     * @param string $leafTenantId
     * @param string $applicationId
     * @param OfferingItem[] $offeringItems
     * @throws ApiException
     * @throws HttpException
     */
    public function enableApplicationOfferingItemsForTenantsChain(
        $rootTenantId,
        $leafTenantId,
        $applicationId,
        array $offeringItems
    )
    {
        if ($leafTenantId === $rootTenantId) {
            return;
        }

        // enable only OIs with status 1, other will be ignored
        $neededOfferingItems = static::getOfferingItemsHashTable(
            $this->getActiveOfferingItems($offeringItems)
        );
        $availableOfferingItems = static::getOfferingItemsHashTable(
            $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
        );

        // Enable not available offering items for the parent
        $notAvailableOfferingItems = array_diff_key($neededOfferingItems, $availableOfferingItems);
        if (!empty($notAvailableOfferingItems)) {
            $tenant = $this->getTenant($leafTenantId);
            $this->enableApplicationOfferingItemsForTenantsChain(
                $rootTenantId, $tenant->getParentId(), $applicationId, $notAvailableOfferingItems
            );
            $availableOfferingItems = static::getOfferingItemsHashTable(
                $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
            );
        }

        // Exist if all needed offering items are enabled
        $enabledOfferingItems = array_filter($availableOfferingItems, function ($offeringItem) {
            /** @var OfferingItem $offeringItem */
            return $offeringItem->getStatus();
        });
        $disabledOfferingItems = array_diff_key($neededOfferingItems, $enabledOfferingItems);
        if (empty($disabledOfferingItems)) {
            return;
        }

        // Enable disabled offering items
        $offeringItemsToUpdate = [];
        foreach ($disabledOfferingItems as $offeringItemHash => $neededOfferingItem) {
            /** @var OfferingItem $neededOfferingItem */

            $oi = new OfferingItem();
            $oi->setApplicationId($applicationId);
            $oi->setName($neededOfferingItem->getName());
            $oi->setStatus(static::OFFERING_ITEM_STATUS_ACTIVE);

            if ($neededOfferingItem->hasInfraId()) {
                $oi->setInfraId($neededOfferingItem->getInfraId());
            }

            $offeringItemsToUpdate[] = $oi;
        }

        // Sent request to update offering items
        $this->updateTenantApplicationOfferingItems($leafTenantId, $offeringItemsToUpdate);
    }

    /**
     * @param $applicationType
     * @return Application|null
     */
    public function getApplication($applicationType)
    {
        return $this->memoize(function () use ($applicationType) {
            $application = $this->getApplicationByType($applicationType);
            if (is_null($application)) {
                throw new RestException(404, Str::format(
                    'Cannot fetch application "%s".',
                    $applicationType
                ));
            }

            return $application;
        }, $applicationType);
    }

    public function getOfferingItemsAvailableForChildHashMap($userPersonalTenantId)
    {
        $result = [];
        foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $result += $this->getApplicationOfferingItemsAvailableForChildHashMap($applicationType, $userPersonalTenantId);
        }

        return $result;
    }

    /**
     * Update specified only offering items for $leafTenantId.
     * Recursive enable offering items for tenant's parent (see method enableUnlimitedOfferingItemsForTenantsChain)
     *
     * @param string $rootTenantId
     * @param string $leafTenantId
     * @param string $applicationId
     * @param OfferingItem[] $offeringItems
     * @param string|null $targetEdition
     * @throws ApiException
     * @throws HttpException
     */
    public function updateApplicationOfferingItemsForTenantsChain(
        $rootTenantId,
        $leafTenantId,
        $applicationId,
        array $offeringItems,
        $targetEdition = null
    )
    {
        if ($leafTenantId === $rootTenantId) {
            throw new HttpException(Str::format(
                'Unable to update offers for yourself. Root tenant ID equals leaf tenant ID %s.',
                $leafTenantId
            ), 403);
        }

        $neededOfferingItems = static::getOfferingItemsHashTable($offeringItems);
        $availableOfferingItems = static::getOfferingItemsHashTable(
            $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
        );

        // enable not available offering items for tenant's parent if there are such
        $notAvailableOfferingItems = array_diff_key(
            $this->getActiveOfferingItems($neededOfferingItems),
            $availableOfferingItems
        );
        if (!empty($notAvailableOfferingItems)) {
            $tenant = $this->getTenant($leafTenantId);
            $this->enableApplicationOfferingItemsForTenantsChain(
                $rootTenantId, $tenant->getParentId(), $applicationId, $notAvailableOfferingItems
            );

            // refresh a list of available offering items
            $availableOfferingItems = static::getOfferingItemsHashTable(
                $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
            );
        }

        $offeringItemsToUpdate = [];
        foreach ($neededOfferingItems as $offeringItemHash => $offeringItem) {
            /** @var OfferingItem $offeringItem */
            $status = $offeringItem->hasStatus()
                ? $offeringItem->getStatus()
                : static::OFFERING_ITEM_STATUS_ACTIVE;

            $oi = new OfferingItem();
            $oi->setStatus($status);
            $oi->setApplicationId($applicationId);
            $oi->setName($offeringItem->getName());

            if ($offeringItem->hasInfraId()) {
                $oi->setInfraId($offeringItem->getInfraId());
            }

            if ($status !== static::OFFERING_ITEM_STATUS_INACTIVE && $offeringItem->hasQuota()) {
                $offeringItemQuota = $offeringItem->getQuota();

                $quota = new Quota();
                $quota->setValue($offeringItemQuota->getValue());
                $quota->setOverage($offeringItemQuota->getOverage());

                $quota->setVersion(0);

                // set current version of quota
                /** @var OfferingItem $availableOfferingItem */
                $availableOfferingItem = Arr::get($availableOfferingItems, $offeringItemHash);

                if ($availableOfferingItem && $availableOfferingItem->hasQuota()) {
                    $availableOfferingItemQuota = $availableOfferingItem->getQuota();
                    if ($availableOfferingItemQuota->hasVersion()) {
                        $quota->setVersion($availableOfferingItemQuota->getVersion());
                    }
                }

                $oi->setQuota($quota);
            }

            $offeringItemsToUpdate[] = $oi;
        }

        $usersOisAfterSwitch = [];
        if ($targetEdition !== null) {
            $tenant = $this->getTenant($leafTenantId);
            if ($tenant->getKind() === 'customer') {
                $this->switchTenantEdition($leafTenantId, $applicationId, $targetEdition);
                foreach ($this->getTenantUsers($leafTenantId) as $userId) {
                    $user = $this->getUser($userId);
                    $personalTenantId = $user->getPersonalTenantId();
                    if (is_array($personalTenantId)) {
                        $personalTenantId = reset($personalTenantId);
                    }
                    $usersOisAfterSwitch[$personalTenantId] =
                        $this->getTenantApplicationOfferingItems($personalTenantId, $applicationId);
                }
            }
        }
        // sent request to update offering items
        $this->updateTenantApplicationOfferingItems($leafTenantId, $offeringItemsToUpdate);
        return $usersOisAfterSwitch;
    }

    /* ====== Methods to manage reports ====== */

    /**
     * @param string $reportId
     * @return StoredReportParamsItems[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getStoredUsageReports($reportId)
    {
        return $this->authorizedCall(function () use ($reportId) {
            return $this->getReportsApi()->getReportsStoredByReportId($reportId);
        })->getItems();
    }

    /**
     * @param string $reportId
     * @param string $storedReportId
     * @param string $destinationFile
     * @throws ApiException
     * @throws HttpException
     */
    public function downloadStoredUsageReportArchive($reportId, $storedReportId, $destinationFile)
    {
        $this->authorizedCall(function () use ($reportId, $storedReportId, $destinationFile) {
            $resourcePath = Str::format('/reports/%s/stored/%s', $reportId, $storedReportId);
            $apiClientConfig = $this->getApiClient()->getConfig();
            $url = $apiClientConfig->getHost() . $resourcePath;
            $headers = [
                'Authorization' => 'Bearer ' . $apiClientConfig->getAccessToken(),
                'User-Agent' => $apiClientConfig->getUserAgent(),
            ];
            $this->getLogger()->info(Str::format(
                'Download file "%s" to "%s".',
                $url, $destinationFile
            ));
            $this->downloadAndSaveFile($url, $destinationFile, $headers);
        });
    }

    /**
     * @param string $tenantId
     * @return Report
     * @throws ApiException
     * @throws HttpException
     */
    public function orderCurrentUsageReportForAccounts($tenantId)
    {
        $params = new ReportPostParameters();
        $params->setTenantId($tenantId);
        $params->setKind(ReportPostKind::CURRENT);
        $params->setLevel(ReportPostLevel::ACCOUNTS);

        $schedule = new ReportPostSchedule();
        $schedule->setType(ReportPostType::ONCE);

        $body = new ReportPost();
        $body->setParameters($params);
        $body->setSchedule($schedule);
        $body->setRecipients([]);
        $body->setResultAction(ReportPostResultAction::SAVE);

        return $this->authorizedCall(function () use ($body) {
            return $this->getReportsApi()->postReports($body);
        });
    }

    /* ====== Methods to manage infra objects ====== */

    /**
     * @param string $tenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantInfrasUuids($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsInfraByTenantId($tenantId);
        })->getInfras();
    }

    /**
     * @param string[] $uuids
     * @return Infra[]
     * @throws ApiException
     * @throws HttpException
     */
    public function fetchInfras($uuids)
    {
        return $this->authorizedCall(function () use ($uuids) {
            return $this->getInfraApi()->getInfra($uuids);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @param string $capability
     * @return Infra[]
     * @throws ApiException
     * @throws HttpException
     */
    public function fetchTenantInfras($tenantId, $capability = null)
    {
        $uuids = $this->getTenantInfrasUuids($tenantId);
        if (empty($uuids)) {
            return [];
        }
        $infras = $this->fetchInfras($uuids);
        if ($capability) {
            $infras = array_filter($infras, function ($infra) use ($capability) {
                return in_array($capability, $infra->getCapabilities());
            });
        }

        return $infras;
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function getInfrasUuidsForTenantsChain($rootTenantId, $leafTenantId)
    {
        $tenantPath = $this->resolveTenantPath($rootTenantId, $leafTenantId);

        $infras = [];
        foreach ($tenantPath as $tenantId) {
            $infras = array_merge($infras, $this->getTenantInfrasUuids($tenantId));
        }

        return array_unique($infras);
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return Infra[]
     * @throws ApiException
     * @throws HttpException
     */
    public function fetchInfrasForTenantsChain($rootTenantId, $leafTenantId)
    {
        $uuids = $this->getInfrasUuidsForTenantsChain($rootTenantId, $leafTenantId);

        return $this->fetchInfras($uuids);
    }

    /* ====== Protected utility methods ====== */

    /**
     * @return string
     * @throws HttpException
     * @throws IOException
     */
    protected function getServerUrl()
    {
        if (!$this->serverUrl) {
            $this->serverUrl = $this->resolveServerUrlForLogin();
            $this->serverUrl .= static::URL_PATH_API_V2;
        }

        return $this->serverUrl;
    }

    /* ====== Private methods ====== */

    /**
     * @return string
     * @throws HttpException
     * @throws IOException
     */
    private function resolveServerUrlForLogin()
    {
        return $this->fromCache(function () {
            $accountInfo = $this->fetchAccountInfo($this->login);

            return rtrim($accountInfo->server_url, '/');
        }, static::CACHE_TTL, false, $this->getCredentialsHash());
    }

    /**
     * @param $userPersonalTenantId
     * @return OfferingItemOutput[]
     */
    public function getApplicationOfferingItemsAvailableForChildHashMap($applicationType, $userPersonalTenantId)
    {
        return $this->memoize(function () use ($userPersonalTenantId, $applicationType) {
            $application = $this->getApplication($applicationType);
            $applicationId = $application->getId();

            $offeringItemsAvailableForChild = $this->getOfferingItemsAvailableForChild($userPersonalTenantId, 'unit');
            $availableOfferingItems = array_filter($offeringItemsAvailableForChild, function ($oi) use ($applicationId) {
                /** @var \Acronis\Cloud\Client\Model\Tenants\OfferingItem $oi */
                return $oi->getApplicationId() === $applicationId;
            });
            $availableOfferingItemsHashmap = Arr::map($availableOfferingItems, function ($item) {
                /** @var OfferingItemOutput $item */
                return $item->getName();
            }, function ($item) {
                return $item;
            }, false);

            return $availableOfferingItemsHashmap;
        }, $userPersonalTenantId . '-' . $applicationType);
    }

    /**
     * @param string $login
     * @return object
     * @throws HttpException
     * @throws IOException
     */
    private function fetchAccountInfo($login)
    {
        $url = rtrim($this->url, '/') . static::URL_PATH_API_V1;

        $apiClient = $this->createApiClient($url);

        $path = '/accounts';
        $method = 'GET';
        $queryParameters = ['login' => $login];
        $body = null;
        $headers = [];

        list($response, $statusCode, $httpHeader) = $apiClient->callApi(
            $path, $method, $queryParameters, $body, $headers
        );

        return $response;
    }

    /**
     * @param string $url
     * @param string $destinationFile
     * @param array $headers
     * @throws IOException
     * @throws HttpException
     */
    private function downloadAndSaveFile($url, $destinationFile, $headers)
    {
        $client = new Client();

        $fp = fopen($destinationFile, 'wb');
        try {
            $client->request('GET', $url, [
                'sink' => $fp,
                'headers' => $headers,
                'verify' => false,
                'decode_content' => false
            ]);
        } catch (BadResponseException $e) {
            throw new HttpException($e->getMessage(), $e->getCode());
        } catch (GuzzleException $e) {
            throw new IOException($e->getMessage(), $e->getCode());
        }
    }

    protected function getCredentialsHash()
    {
        return md5($this->getUrl() . $this->getLogin() . $this->getPassword());
    }

    /**
     * @param OfferingItem[] $offeringItems
     * @return array
     */
    private function getActiveOfferingItems(array $offeringItems)
    {
        return array_filter($offeringItems, function ($offeringItem) {
            /** @var OfferingItem $offeringItem */
            // if status is not specified the offering item will be enabled
            if (!$offeringItem->hasStatus()) {
                return true;
            }

            return $offeringItem->getStatus() === static::OFFERING_ITEM_STATUS_ACTIVE;
        });
    }
}
