<?php
/**
 * @Copyright © 2002-2021 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Console\Command\UsageReport;

use Acronis\App\Console\Command\LockableTrait;
use Acronis\App\Console\Command\Loggable;
use Acronis\App\Database\Migrations\InitialMigration;
use Acronis\App\Services\Config\ConfigAwareTrait;
use Acronis\App\Services\Config\ConfigSchema;
use Acronis\App\Services\Database\DatabaseAwareTrait;
use Acronis\App\Services\Redis\RedisAwareTrait;
use Acronis\App\Services\UsageReportManager\UsageReportManagerAwareTrait;
use Acronis\App\UsageReport\TenantsMap\TenantsMap;
use Acronis\App\Utils\CacheableTrait;
use Arr;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;

final class FlushAll extends Loggable
{
    use CacheableTrait;
    use ConfigAwareTrait;
    use DatabaseAwareTrait;
    use LockableTrait;
    use RedisAwareTrait;
    use UsageReportManagerAwareTrait;

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $this->flushReportsTable();
        $this->cleanReportDownloadPath();
        $this->invalidateReportData();
        $this->releaseLocks();
    }

    protected function getCommandName()
    {
        return 'usage-report:flushall';
    }

    protected function getCommandDescription()
    {
        return 'Command for usage report flushing.';
    }

    protected function getCommandHelp()
    {
        return <<<TEXT
<info>
This command drops report table and creates a new one from migration, removes all 
downloaded reports, invalidates all report cache and storage from Redis.
</info>
TEXT;
    }

    private function releaseLocks()
    {
        $this->getLog()->notice('Releasing all report usage-report:process command locks.');
        $this->releaseAllLocks(Process::COMMAND_NAME);
    }

    private function flushReportsTable()
    {
        $migration = new InitialMigration();
        $this->getLog()->notice('Dropping reports table.');
        $migration->down($this->getDatabaseConnection()->getPdo());
        $this->getLog()->notice('Creating reports table.');
        $migration->up($this->getDatabaseConnection()->getPdo());
    }

    private function cleanReportDownloadPath()
    {
        $reportManager = $this->getUsageReportManager();
        $reportManager->cleanReportDownloadPath();
    }

    private function invalidateReportData()
    {
        $redis = $this->getRedis();
        $this->getLog()->notice('Invalidating reports storage (actually Redis).');
        $redis->deleteAll($this->getReportStorageWildcard());
        $this->getLog()->notice('Invalidating reports cache (actually Redis).');
        $redis->deleteAll($this->getReportCacheWildcard());
    }

    private function getReportStorageWildcard()
    {
        return Arr::get(
                $this->getConfig()->getUsageReportConfig(), ConfigSchema::NODE_USAGE_REPORT_PREFIX, ''
        ) . '*';
    }

    private function getReportCacheWildcard()
    {
        return Arr::get(
                $this->getConfig()->getCacheConfig(), ConfigSchema::NODE_CACHE_PREFIX, ''
        ) . $this->getReportStorageWildcard();
    }
}