<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Console\Command\UsageReport;

use Acronis\App\Console\Command\Loggable;
use Acronis\App\Database\Model\ReportEntry;
use Acronis\App\Database\Model\ReportEntryHelper;
use Acronis\App\Services\Apsc\ApscInstanceFactoryTrait;
use Acronis\App\Services\Config\ConfigAwareTrait;
use Acronis\App\Services\Redis\RedisAwareTrait;
use Acronis\App\Services\UsageReportManager\UsageReportManagerAwareTrait;
use Symfony\Component\Console\Helper\Table;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\BufferedOutput;
use Symfony\Component\Console\Output\OutputInterface;

final class ViewList extends Loggable
{
    use UsageReportManagerAwareTrait;
    use RedisAwareTrait;
    use ConfigAwareTrait;
    use ApscInstanceFactoryTrait;

    const PROPERTY_DATACENTER_ID = 'datacenter_id';
    const PROPERTY_DATE = 'date';
    const PROPERTY_ID = 'id';

    const SYNTHETIC_PROPERTY_DC_ID = 'dc_id';
    const SYNTHETIC_PROPERTY_DC_NAME = 'dc_name';
    const SYNTHETIC_PROPERTY_FILE = 'file_path';
    const SYNTHETIC_PROPERTY_STATUS = 'status';

    const PROPERTY_NOT_RESOLVED = '----';

    private $datacenters = [];

    private static $columns = [
        self::PROPERTY_ID => 'ID',
        self::PROPERTY_DATE => 'Date',
        self::PROPERTY_DATACENTER_ID => 'Datacenter UUID',
        self::SYNTHETIC_PROPERTY_DC_NAME => 'Datacenter name',
        self::SYNTHETIC_PROPERTY_DC_ID => 'Datacenter ID',
        self::SYNTHETIC_PROPERTY_STATUS => 'Report status',
        self::SYNTHETIC_PROPERTY_FILE => 'File path',
    ];

    /**
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int|null|void
     * @throws \Exception
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $helper = new ReportEntryHelper();
        $reports = $helper->getAllReports();

        if (!$reports->count()) {
            $output->writeln('There are no reports to be processed.');
            return;
        }

        $output->writeln('List of reports:');

        $tableOutput = new BufferedOutput();
        $table = $this->buildConsoleTable($reports, $tableOutput);
        $table->render();

        $output->writeln($tableOutput->fetch());
    }

    protected function getCommandName()
    {
        return 'usage-report:view-list';
    }

    protected function getCommandDescription()
    {
        return 'Command for usage report processing.';
    }

    protected function getCommandHelp()
    {
        return <<<TEXT
<info>
This command drops report table at creates a new one from migration, removes all 
downloaded reports, invalidates all report cache and storage from Redis
</info>
TEXT;
    }

    /**
     * @param $reports
     * @param $output
     * @return Table
     * @throws \Exception
     */
    private function buildConsoleTable($reports, $output)
    {
        $table = new Table($output);
        $table->setHeaders(array_values(static::$columns));
        foreach ($reports as $report) {
            $table->addRow($this->buildReportRow($report));
        }

        return $table;
    }

    /**
     * @param ReportEntry $report
     * @throws \Exception
     * @return string[]
     */
    private function buildReportRow($report)
    {
        $row = [];
        foreach (array_keys(static::$columns) as $property) {
            switch ($property) {
                case self::SYNTHETIC_PROPERTY_STATUS:
                    $value = ReportEntry::getStatusDescription($report->status);
                    break;
                case self::SYNTHETIC_PROPERTY_FILE:
                    $value = $report->file_path;
                    break;
                case self::SYNTHETIC_PROPERTY_DC_NAME:
                    $dc = $this->getDatacenter($report->datacenter_id, $report->aps_instance_id);
                    $value = is_null($dc) ? static::PROPERTY_NOT_RESOLVED : $dc->name;
                    break;
                case self::SYNTHETIC_PROPERTY_DC_ID:
                    $dc = $this->getDatacenter($report->datacenter_id, $report->aps_instance_id);
                    $value = is_null($dc) ? static::PROPERTY_NOT_RESOLVED : $dc->id;
                    break;
                default:
                    $value = $report->{$property};
            }

            $row[] = $value;
        }

        return $row;
    }

    /**
     * @param $datacenterId
     * @param $apsInstanceId
     * @return \datacenter|null
     * @throws \Exception
     */
    private function getDatacenter($datacenterId, $apsInstanceId)
    {
        if (!isset($this->datacenters[$datacenterId])) {
            $this->datacenters[$datacenterId] = $this->safeFetchDatacenter($datacenterId, $apsInstanceId);
        }

        return $this->datacenters[$datacenterId];
    }

    /**
     * @param $datacenterId
     * @param $apsInstanceId
     * @return \datacenter|null
     * @throws \Exception
     */
    private function safeFetchDatacenter($datacenterId, $apsInstanceId)
    {
        $rqlFilter = \RqlBuilder::createInstance()
            ->implementing(\APSResource::CLOUD_BACKUP_DATACENTER_TYPE_MAJOR)
            ->eq('aps.id', $datacenterId)
            ->getQuery();
        $apsc = $this->createApscInstance($apsInstanceId);

        // We a suppressing an APS Controller exception because we should be able to list reports anyway.
        // In case when error occurs, we won't see only datacenter name and numeric APS id.
        try {
            $dcs = $apsc->getResources($rqlFilter);
            return reset($dcs) ?: null;
        } catch (\Exception $e) {
            return null;
        }
    }
}