<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\ApplicationsManager;

class Model_MspGroupManager implements Model_MspGroupManagerInterface
{
    const LOG_EVENT_METHOD_WAS_CALLED = 'Method "%s" was called.';
    const LOG_EVENT_MIGRATE_GROUP = 'Migrate group with ID :group_id was updated. Details: group_old_name=":group_old_name", group_new_name=":group_new_name", group_new_version=:group_new_version, old_parent_id=:old_parent_id, new_parent_id=:new_parent_id';
    const LOG_EVENT_NORMALIZE_GROUP_NAME_AFTER_MIGRATE = 'Name of group with ID :group_id was normalized after migration. Details: group_old_name=":group_old_name", group_new_name=":group_new_name", group_new_version=:group_new_version';

    const RESELLER_GROUP_NAME_TEMPLATE = '{companyName} (Reseller {accountId})';
    const CUSTOMER_GROUP_NAME_TEMPLATE = '{companyName} /{suffix}';
    const CUSTOMER_GROUP_NAME_PATTERN = '/^%s(?: \/(\d+))?$/i';
    const PATTERN_DELIMITER = '/';

    const EXCEPTION_CANNOT_FIND_GROUP_MESSAGE = "Cannot find group by name \"%s\"";
    const EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE = "Cannot create group for account \"%s\"";
    const WARNING_INTERNAL_TAG_WAS_NOT_UPDATED = 'Internal tag was not updated';

    const HASH = '#';

    const UNLIMITED_QUOTA_VALUE = -1;

    const OSA_ACCOUNT_ID = 'osa_account_id';
    const OSA_SUBSCRIPTION_ID = 'osa_subscription_id';
    const TIER_LEVEL = 'aps-abc-tier';
    const START_PROD_DATE = 'aps-abc-start-prod-date';

    /** @var Service_MspApiClient_ApiInterface */
    private $_mspApi;
    /** @var Service_APSApi_ApiInterface */
    private $_apsApi;

    /**
     * Model_MspGroupManager constructor.
     *
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param Service_APSApi_ApiInterface $apsApi
     */
    public function __construct(
        Service_MspApiClient_ApiInterface $mspApi,
        Service_APSApi_ApiInterface $apsApi
    )
    {
        $this->_mspApi = $mspApi;
        $this->_apsApi = $apsApi;
    }

    /**
     * @param tenant $tenant
     * @param int $parentGroupId
     * @param int $datacenterId
     * @param $tenantStrategy
     * @return int
     * @throws Service_BaseClient_HTTPException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     * @throws \Rest\RestException
     */
    public function createGroupForReseller($tenant, $parentGroupId, $datacenterId = null, $tenantStrategy = null)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupName = self::formatResellerGroupName(
            $tenant->account->companyName,
            $tenant->account->id
        );

        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();
        $groupInfo->setName($groupName);
        $groupInfo->setKind(Service_MspApiClient_Request_Struct_CreateGroupInterface::RESELLER_GROUP_KIND_ID);

        $pricing = $groupInfo->getPricing();
        $pricing->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $groupInfo->setCustomerId(Arr::encode([
            self::OSA_ACCOUNT_ID => $tenant->account->id,
            self::OSA_SUBSCRIPTION_ID => $tenant->subscription->subscriptionId,
        ]));

        $this->initGroupInfoContact($tenant, $groupInfo);

        try {
            $createResult = $this->createGroup($parentGroupId, $groupInfo);
        } catch (Exception $ex) {
            if ($ex->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_CONFLICT) {
                throw $ex;
            }

            $group = $this->findGroupByName($parentGroupId, $groupName);
            if (!is_null($group)) {
                $groupId = $group->getId();
                $this->_apsApi->checkGroupCanBeAssignedOrEx($groupId, $datacenterId);
                $this->updateGroupForReseller($groupId, $tenant);
                return $groupId;
            }

            throw new Exception(sprintf(self::EXCEPTION_CANNOT_FIND_GROUP_MESSAGE, $groupName), 0, $ex);
        }

        $version = $createResult->getVersion();

        if (!empty($tenant->startProductionTime) || !$tenant->minimalCommitmentTier->isEmpty()) {
            $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
            $minimalCommitmentTier = $tenant->minimalCommitmentTier;
            $internalTag = [];
            if (!$tenant->minimalCommitmentTier->isEmpty()) {
                $internalTag[self::TIER_LEVEL] = $minimalCommitmentTier->level;
            }
            if (!empty($tenant->startProductionTime)) {
                $internalTag[self::START_PROD_DATE] = $tenant->startProductionTime;
            }
            $updateGroupInfo->setInternalTag(Arr::encode($internalTag));
            $this->updateGroup($createResult->getId(), $updateGroupInfo, $version);
        }

        $applicationManager = $this->createApplicationsManager($tenant, $createResult->getId(), $tenantStrategy);

        try {
            // it's a w/a to disable all offering items after creating via APIv1
            $applicationManager->deactivateAllTenantOfferingItems();
        } catch (Exception $e) {
            Log::getInstance()->warning(
                'Unable to sanitize offering items for tenant "%s". Error: "%s".',
                $createResult->getId(), $e->getMessage()
            );
        }

        $applicationManager->updateOfferingItems();

        return $createResult->getId();
    }

    /**
     * @param tenant $tenant
     * @param int $parentGroupId
     * @param string $folderName
     * @return int
     * @throws Service_BaseClient_HTTPException
     */
    public function createOrFindFolder($tenant, $parentGroupId, $folderName)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupResult = $this->findGroupByName($parentGroupId, $folderName);

        if (!empty($groupResult)) {
            return $groupResult->getId();
        }

        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();
        $groupInfo->setName($folderName);
        $groupInfo->setKind(Service_MspApiClient_Request_Struct_CreateGroupInterface::FOLDER_GROUP_KIND_ID);
        $groupInfo->getPricing()->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        //TODO: set production mode here
        $this->initGroupInfoUnlimitedPrivileges($groupInfo);

        try {
            $createResult = $this->createGroup($parentGroupId, $groupInfo);
        } catch (Exception $ex) {
            if ($ex->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_CONFLICT) {
                throw $ex;
            }

            $createResult = $this->findGroupByName($parentGroupId, $folderName);
            if (is_null($createResult)) {
                throw new Exception(sprintf(self::EXCEPTION_CANNOT_FIND_GROUP_MESSAGE, $folderName), 0, $ex);
            }
        }

        return $createResult->getId();
    }

    /**
     * @param $groupId
     * @param $tenant
     * @param $tenantStrategy
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     * @throws \Rest\RestException
     */
    public function updateFolderOfferingItems($groupId, $tenant, $tenantStrategy)
    {
        $applicationManager = $this->createApplicationsManager($tenant, $groupId, $tenantStrategy);
        $applicationManager->updateFolderOfferingItems();
    }

    public function initGroupInfoUnlimitedPrivileges(Service_MspApiClient_Request_Struct_GroupInterface $groupInfo)
    {
        $privileges = $groupInfo->getPrivileges();

        $privileges->setMailboxCount(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setMobileCount(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setVmCount(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setWorkstationCount(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setStorageSize(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setLocalStorageSize(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setServerCount(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setWebHostingServerCount(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);
        $privileges->setWebsiteCount(Model_MspGroupManager::UNLIMITED_QUOTA_VALUE);

        // w/a for APSOB-5026 ---
        $privileges->setUniversalDevicesCount(0);
        $privileges->setWinServerEssentialCount(0);
        $privileges->setVirtualHostCount(0);
        // ----------------------

        // w/a for APSOB-5012
        $localBackup = $privileges->getLocalBackup();
        $localBackup->setPermitted(1);
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @return int
     * @throws Exception
     */
    public function updateGroupForReseller($groupId, $tenant)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();

        $groupName = self::formatResellerGroupName(
            $tenant->account->companyName,
            $tenant->account->id
        );
        $updateGroupInfo->setName($groupName);

        $pricing = $updateGroupInfo->getPricing();
        $pricing->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $customerIdParts = Arr::decode($groupInfo->getCustomerId());
        $customerIdParts[self::OSA_ACCOUNT_ID] = $tenant->account->id;
        $customerIdParts[self::OSA_SUBSCRIPTION_ID] = $tenant->subscription->subscriptionId;
        $updateGroupInfo->setCustomerId(Arr::encode($customerIdParts));

        $minimalCommitmentTier = $tenant->minimalCommitmentTier;
        $internalTag = Arr::decode($groupInfo->getInternalTag());
        if (!$minimalCommitmentTier->isEmpty()) {
            $internalTag[self::TIER_LEVEL] = $minimalCommitmentTier->level;
        }
        if (!empty($tenant->startProductionTime)) {
            $internalTag[self::START_PROD_DATE] = $tenant->startProductionTime;
        }
        $updateGroupInfo->setInternalTag(Arr::encode($internalTag));

        $this->initGroupInfoContact($tenant, $updateGroupInfo);

        $version = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        $applicationManager = $this->createApplicationsManager($tenant, $groupInfo->getUuid());
        $applicationManager->updateOfferingItems();

        return $version;
    }

    /**
     * @param tenant $tenant
     * @param int $parentGroupId
     * @param int $groupKind
     * @param string $language
     * @return string
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function createGroupForCustomer(
        $tenant,
        $parentGroupId,
        $groupKind = Service_MspApiClient_Request_Struct_CreateGroup::EUC_GROUP_KIND_ID,
        $language = null
    )
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $companyName = $tenant->account->companyName;

        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();
        $groupInfo->setName($companyName . ' (' . strval($tenant->subscription->subscriptionId) . ')');
        $groupInfo->setKind($groupKind);

        $pricing = $groupInfo->getPricing();
        $mode = $groupKind !== Service_MspApiClient_Request_Struct_CreateGroupInterface::EUC_GROUP_KIND_ID ?
            Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE :
            $this->getGroupMode($tenant);
        $pricing->setMode($mode);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $groupInfo->setCustomerId(Arr::encode([
            self::OSA_ACCOUNT_ID => $tenant->account->id,
            self::OSA_SUBSCRIPTION_ID => $tenant->subscription->subscriptionId,
        ]));

        $this->initGroupInfoContact($tenant, $groupInfo);

        if (!is_null($language)) {
            $groupInfo->setLanguage($language);
        }
        $createResult = $this->createGroup($parentGroupId, $groupInfo);

        // it's a w/a to disable all offering items after creating via APIv1 enabled by default
        try {
            $applicationManager = $this->createApplicationsManager($tenant, $createResult->getId());
            $applicationManager->deactivateAllTenantOfferingItems();
        } catch (Exception $e) {
            Log::getInstance()->warning(
                'Unable to sanitize offering items for tenant "%s". Error: "%s".',
                $createResult->getId(), $e->getMessage()
            );
        }
        // w/a because we still need to resolve numeric ID as we use APIv1 for locking group
        $group = $this->_mspApi->getGroupById($createResult->getId());

        return $group->getId();
    }

    /**
     * @param tenant $tenant
     * @param $tenantStrategy
     * @throws \Rest\RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function updatePrivilegesForCustomers($tenant, $tenantStrategy, $skipOffieringItemValidation = false)
    {
        $applicationManager = $this->createApplicationsManager($tenant, null, $tenantStrategy);
        if ($skipOffieringItemValidation) {
            $applicationManager->enableOfferingItemsValidationSkip();
        }
        return $applicationManager->updateOfferingItems();
    }

    /**
     * @param int $groupId
     * @param int $parentGroupId
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function updateGroupForCustomer($tenant, $newParentGroupId = null)
    {
        $groupId = $tenant->mspGroupId;
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        $companyName = $tenant->account->companyName;

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $updateGroupInfo->setName($companyName . ' (' . strval($tenant->subscription->subscriptionId) . ')');
        if (!is_null($newParentGroupId)) {
            $updateGroupInfo->setParentId($newParentGroupId);
        }

        $pricing = $updateGroupInfo->getPricing();
        $mode = $groupInfo->getKind() !== Service_MspApiClient_Request_Struct_CreateGroupInterface::EUC_GROUP_KIND_ID ?
            Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE :
            $this->getGroupMode($tenant);

        $currentMode = $groupInfo->getPricing()->getMode();
        if ($currentMode !== $mode) {
            if ($currentMode === Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE) {
                // never change pricing mode from production mode
                Log::getInstance()->warning(
                    'The current pricing mode of the tenant is "%s" while the pricing mode of the subscription is "%s".',
                    $currentMode, $mode
                );
            } else {
                $pricing->setMode($mode);
            }
        }

        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $customerIdParts = Arr::decode($groupInfo->getCustomerId());
        $customerIdParts[self::OSA_ACCOUNT_ID] = $tenant->account->id;
        $customerIdParts[self::OSA_SUBSCRIPTION_ID] = $tenant->subscription->subscriptionId;
        $updateGroupInfo->setCustomerId(Arr::encode($customerIdParts));

        $this->initGroupInfoContact($tenant, $updateGroupInfo);

        $version = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        return $version;
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @param int $newParentGroupId
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function migrateGroupForCustomer($groupId, $tenant, $newParentGroupId)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        if ($groupInfo->getParentId() == $newParentGroupId) {
            return $groupInfo->getVersion();
        }

        $companyName = $tenant->account->companyName;

        // Actually, we split moving group in 2 step (2 http requests)

        // 1) Set name that will be unique for old and new parents and move group (first http request)

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $tmpGroupName = $companyName . " ({$groupId})";
        $updateGroupInfo->setName($tmpGroupName);
        $updateGroupInfo->setParentId($newParentGroupId);
        $groupNewVersion = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        Log::getInstance()->event(
            self::LOG_EVENT_MIGRATE_GROUP,
            [
                ':group_id' => $groupInfo->getId(),
                ':group_old_name' => $groupInfo->getName(),
                ':group_new_name' => $tmpGroupName,
                ':old_parent_id' => $groupInfo->getParentId(),
                ':new_parent_id' => $newParentGroupId,
                ':group_new_version' => $groupNewVersion,
            ]
        );

        // 2) Rename moved group deleting unique part (second http request) and enable group if it status is inactive

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $updateGroupInfo->setName($companyName . ' (' . strval($tenant->subscription->subscriptionId) . ')');
        if ($groupInfo->getStatus() == Model_TenantManager::STATUS_INACTIVE) {
            $updateGroupInfo->setStatus(Model_TenantManager::STATUS_ACTIVE);
        }

        $groupNewVersion = $this->updateGroup($groupId, $updateGroupInfo, $groupNewVersion);

        Log::getInstance()->event(
            self::LOG_EVENT_NORMALIZE_GROUP_NAME_AFTER_MIGRATE,
            [
                ':group_id' => $groupInfo->getId(),
                ':group_old_name' => $tmpGroupName,
                ':group_new_name' => $updateGroupInfo->getName(),
            ]
        );

        return $groupNewVersion;
    }

    /**
     * @param string $companyName
     * @param int $id
     * @return string
     */
    static public function formatResellerGroupName($companyName, $id)
    {
        return strtr(self::RESELLER_GROUP_NAME_TEMPLATE, [
            '{companyName}' => $companyName,
            '{accountId}' => $id,
        ]);
    }

    /**
     * @param tenant $tenant
     * @param string|null $tenantId
     * @param $tenantStrategy
     * @return ApplicationsManager
     * @throws \Rest\RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function createApplicationsManager(
        $tenant,
        $tenantId = null,
        $tenantStrategy = null
    )
    {
        return new ApplicationsManager($tenant, $this->_mspApi, $tenantId, $tenantStrategy);
    }

    /**
     * @param int $parentGroupId
     * @param string $companyName
     * @return Service_MspApiClient_Response_Struct_GroupItem
     * @throws Exception
     */
    private function findGroupByName($parentGroupId, $companyName)
    {
        /**
         * @var Service_MspApiClient_Response_Struct_GroupItem[] $groups
         */
        $groups = $this->_mspApi->getGroups($parentGroupId);
        foreach ($groups as $group) {
            if ($group->getName() === $companyName) {
                return $group;
            }
        }

        return null;
    }

    /**
     * @param tenant $tenant
     * @param Service_MspApiClient_Request_Struct_GroupInterface $groupInfo
     */
    private function initGroupInfoContact($tenant, Service_MspApiClient_Request_Struct_GroupInterface $groupInfo)
    {
        $admin = $this->_apsApi->getAdmin($tenant);
        if (is_null($admin)) {
            return;
        }

        $contact = $groupInfo->getContact();
        $contact->setEmail($admin->email);
        $contact->setFirstName($admin->givenName);
        $contact->setLastName($admin->familyName);
        $contact->setPhone(str_replace(self::HASH, '', $admin->telWork));

        $addressPostal = $admin->addressPostal;
        if (is_null($addressPostal)) {
            return;
        }
        $contact->setCountry(strtoupper($addressPostal->countryName));
        $contact->setZipCode($addressPostal->postalCode);
        $contact->setState($addressPostal->region);
        $contact->setCity($addressPostal->locality);
        $contact->setAddress1($addressPostal->streetAddress);
        $contact->setAddress2($addressPostal->extendedAddress);
    }

    /**
     * @param int $parentGroupId
     * @param Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo
     * @return \Acronis\Cloud\Client\Model\Tenants\Tenant
     * @throws Exception
     */
    private function createGroup($parentGroupId, Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo)
    {
        $creationResult = $this->_mspApi->createGroup($groupInfo, $parentGroupId);
        if (!$creationResult) {
            throw new Exception(sprintf(self::EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE, $groupInfo->getName()));
        }

        return $creationResult;
    }

    /**
     * @param $groupId
     * @param $updateGroupInfo
     * @param $version
     * @return int
     */
    private function updateGroup($groupId, $updateGroupInfo, $version)
    {
        $tenantManager = new Model_TenantManager($this->_mspApi);
        return $tenantManager->updateGroup($groupId, $updateGroupInfo, $version);
    }

    /**
     * @param tenant $tenant
     * @return int
     */
    private function getGroupMode($tenant)
    {
        return $tenant->subscription->trial
            ? Service_MspApiClient_Request_Struct_GroupPricing::TRIAL_MODE
            : Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE;
    }
}
