<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\ApplicationCounter;
use Acronis\App\Base\ApplicationsManager;
use Acronis\App\Base\GroupIdConverterTrait;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Console\Command\LockableTrait;
use Acronis\App\Folders\Cache as FolderCache;
use Acronis\App\Folders\FolderKey;
use Acronis\App\Services\Config\ConfigAwareTrait;
use Acronis\App\Services\Config\ConfigSchema;
use Acronis\App\Services\CountersManager\CountersManager;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Services\Redis\RedisAwareTrait;
use Model\Resource\Tenant\CloudDoesNotMatchWithOsaException;
use Monolog\Logger;

/**
 * Class Model_Resource_Tenant_Customer
 */
class Model_Resource_Tenant_Customer extends Model_Resource_Tenant
{
    use ConfigAwareTrait,
        CountersManagerAwareTrait,
        LockableTrait,
        RedisAwareTrait,
        GroupIdConverterTrait;

    const DEFAULT_FOLDER_NAME = 'Other';

    /** @var FolderCache  */
    private $folderNameToParentIdCache;

    /* ======= RESOURCE METHODS ======= */

    public function __construct(
        $tenant,
        \APS\ControllerProxy $apsc,
        FacadeInterface $facade,
        Model_ManagerFactoryInterface $managerFactory)
    {
        parent::__construct($tenant, $apsc, $facade, $managerFactory);

        $this->folderNameToParentIdCache = new FolderCache();
    }

    public function provision()
    {
        try {
            Log::getInstance()->event(Str::format(
                self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
            ));

            $this->checkProfile();
            $this->linkProfile(false);

            $tenant = $this->tenant;
            $tenant->datacenterId = $this->resolveDatacenterId();
            $this->linkDatacenter(false);

            $this->updateCounterLimits(true);

            if ($tenant->ismigration && !is_null($tenant->mspGroupId)) {
                Log::getInstance()->event(self::LOG_MIGRATION_WAS_STARTED);
                $this->_migrateMspGroup();

                $this->_retrieveUsageForCustomer();
                $this->_updateGroupForCustomer();
            } else {
                try {
                    $tenant->mspGroupId = $this->_createGroupForCustomer();
                } catch (\Exception $e) {
                    if (file_exists(__DIR__ . '/existing_tenants.json')) {
                        $map = json_decode(file_get_contents(__DIR__ . '/existing_tenants.json'), true);
                        $tenantInfo = Arr::get($map, $tenant->account->companyName);
                    }

                    if (isset($tenantInfo) && $tenantInfo !== null) {
                        $tenant->mspGroupId = $tenantInfo['group_id'];
                        $tenant->tenantId = $tenantInfo['tenant_id'];
                    } else {
                        throw $e;
                    }
                }
            }
            $this->updateTenantId();
        } catch (\Exception $e) {
            if (isset($tenantInfo)) {
                Log::getInstance()->notice(Logger::ERROR, sprintf(
                    'Error occurred during resubmit of order for existing Cloud tenant "%s" (sync phase). Error: "%s".',
                    json_encode($this->tenant), $e->getMessage()
                ));
            }

            throw $e;
        }
    }

    public function provisionAsync()
    {
        try {
            Log::getInstance()->event(Str::format(
                self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
            ));
            $this->updateStartProductionTime();

            $this->linkDatacenter(false);
            $this->linkProfile(false);
            $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);
            $this->_subscribeToOnSubscriptionChangeEvent();
            $this->_subscribeToOnUserChangeEvent();
            $this->_subscribeToOnAccountChangeEvent();
        } catch (\Exception $e) {
            if (file_exists(__DIR__ . '/existing_tenants.json')) {
                $map = json_decode(file_get_contents(__DIR__ . '/existing_tenants.json'), true);
                $tenantInfo = Arr::get($map, $this->tenant->account->companyName);
                if ($tenantInfo !== null) {
                    Log::getInstance()->notice(Logger::ERROR, sprintf(
                        'Error occurred during resubmit of order for existing Cloud tenant "%s" (async phase). Tenant info used: "%s". Error: "%s".',
                        json_encode($this->tenant), json_encode($tenantInfo), $e->getMessage()
                    ));
                }
            }

            throw $e;
        }
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->updateCounterLimits();
        $this->_retrieveUsageForCustomer();
    }

    public function configure($new)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateTenant();
    }

    private function updateTenant($updateLegacy = false)
    {
        /**
         * Default lock duration is 300 seconds.
         * Most critical situation is kill of process - in this certain case tenant won't be updated next 5 minutes
         * In case of normal execution and any exception lock will be released automatically
         */
        $tenant = $this->tenant;
        if ($updateLegacy) {
            $this->updateTenantLegacyLogic();
        }

        // todo remove when remove APIv1
        if (is_null($tenant->mspGroupId)) {
            Log::getInstance()->notice(
                'Skipped tenant "%s" update because mspGroupId is not set.',
                $this->tenant->aps->id
            );
            return;
        }

        // 1. update tenant in Cloud
        $this->_updateGroupForCustomer();
        // todo why this check is not placed inside method?
        $editionSwitchedTo = null;
        $usersOis = [];
        if ($this->isActiveSubscription()) {
            // 2. update tenant quotas in Cloud
            $applicationManager = new ApplicationsManager($tenant, $this->getMspApi(), $this->tenant->tenantId, $this);

            $sourceEditions = $applicationManager->getTenantEditions(['backup']);

            $usersOis = $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

            $targetEditions = $applicationManager->getTenantEditions(['backup']);
            $diff = array_diff($targetEditions, $sourceEditions);
            $editionSwitchedTo = (count($diff) && $this->isPackEdition($diff[0])) > 0 ? $diff[0] : null;
        }

        // 3. Get Cloud quotas for users and update all users in order to rely on Cloud quota acquisition logic
        $this->obtainUsersLimitsFromCloud($editionSwitchedTo, $editionSwitchedTo !== null ? $usersOis : null);
    }

    private function isPackEdition(?string $edition): bool
    {
        return $edition !== null && str_starts_with($edition, 'pck_per_');
    }

    private function isActiveSubscription()
    {
        return !$this->tenant->subscription->disabled && $this->tenant->aps->status !== Service_APSApi_Api::APS_SUSPENDED;
    }

    private function updateTenantLegacyLogic()
    {
        /**
         * Todo dump data before and after execution of this method in regress autotests
         * in order to determine if this method make sense
         */

        $needUpdate = $this->updateStartProductionTime();
        $needUpdate |= $this->updateCounterLimits();

        if ($needUpdate && $this->isActiveSubscription()) {
            $this->update();
        }

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            Log::getInstance()->event(self::LOG_UNPROVISIONING_TENANT_IS_MIGRATION);
            return;
        }

        if (is_null($tenant->mspGroupId)) {
            Log::getInstance()->warning(self::LOG_WARNING_UNPROVISION_TENANT_DOES_NOT_HAVE_GROUP);
            return;
        }
        Log::getInstance()->event(sprintf(self::LOG_FORMAT_DELETE_MSP_GROUP, $tenant->account->companyName));

        $this->linkDatacenter(false);

        try {
            $this->_deleteGroupForCustomer();
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                throw $e;
            }
            sleep(10);
            $this->_deleteGroupForCustomer();
        }
    }


    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->datacenterId)) {
            $tenant->datacenterId = globals::DEFAULT_DATACENTER_ID;
        }

        $this->updateStartProductionTime();
        $this->updateTenantId();
        $this->update();

        $realLink = !$tenant->subscription->disabled;
        if ($realLink) {
            $this->checkProfile();
        }

        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);
        $this->_getOrCreateParentGroupId($tenant->datacenter);  // add parents in map
        $this->_updateGroupForCustomer();
        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this, true);
        $this->_upgradeUsers();

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    /* ======= RESOURCE EVENTS ======= */

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();

        if ($this->tenant->ismigration) {
            return;
        }

        $applicationConfig = $this->getApplicationConfig();
        if (!$applicationConfig->service_auto_assignation) {
            return;
        }
        $admin = $this->getApsApi()->getAdmin($this->tenant);
        if (is_null($admin)) {
            return;
        }
        $this->provisionDefaultUser($admin);
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateTenant(true);
    }

    public function onUserChanged($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    public function onAccountChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @return array
     * @throws Exception
     */
    public function getStorages()
    {
        Log::getInstance()->event(
            self::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        return $this->memoize(function () {
            $tenant = $this->tenant;
            $response = [];
            $storages = $this->getMspApi()->getStorages($tenant->mspGroupId);
            if (is_null($storages)) {
                Log::getInstance()->error(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
                throw new Exception(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
            }

            foreach ($storages as $storage) {
                /** @var Service_MspApiClient_Response_Struct_Storage $storage */
                if ($storage->getInherited()) {
                    continue;
                }
                $response[] = [
                    self::FIELD_STORAGE_ID => (string)$storage->getId(),
                    self::FIELD_STORAGE_NAME => $storage->getName(),
                ];
            }

            return $response;
        });
    }

    /**
     * @return object
     */
    public function getDatacenter()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkDatacenter($realLink);
        return (object)$tenant->datacenter->toArray();
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     */
    public function loginExists($login, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        // Is the user's parent the same tenant?
        $user = $this->findUserByLogin($this->tenant->mspGroupId, $login);
        if (empty($user)) {
            // Does the user exist in Cloud
            return $this->checkLoginExist($login, $datacenterId);
        }

        // Is the user already related to an OA user?
        $rql = RqlBuilder::createInstance()
            ->implementing(APSResource::CLOUD_BACKUP_USERS_TYPE)
            ->eq('endUserLogin', $login)
            ->limit(0, 1)
            ->getQuery();
        $resources = $this->getApsc()->impersonate($this->tenant->global)->getResources($rql);
        return !empty($resources);
    }

    /**
     * @return array
     */
    public function getQuotasStatistics()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $quotasStatistics = array();
        $tenant = $this->tenant;
        foreach ($this->fetchSubscriptionCounterNames() as $counterName) {
            $allocatedTotal = 0;
            foreach ($tenant->users as $user) {
                /** @var users $user */
                /** @var UserOfferingItem|null $offeringItem */
                $offeringItem = $this->getUserOfferingItemByCounterName($user, $counterName);

                if (is_null($offeringItem) || !isset($offeringItem->limit)) {
                    continue;
                }
                if ($offeringItem->limit == self::UNLIMITED_VALUE) {
                    $allocatedTotal = self::UNLIMITED_VALUE;
                    break;
                }
                $allocatedTotal += $offeringItem->limit;
            }

            $limit = isset($tenant->{$counterName}->limit) ? $tenant->{$counterName}->limit : 0;

            $statistics = compact(
                'allocatedTotal',
                'limit'
            );

            // For backward compatibility only
            $quotasStatistics[lcfirst($counterName)] = $statistics;
            $quotasStatistics[$counterName] = $statistics;
        }
        return $quotasStatistics;
    }

    public function updateUsageAndLimits()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        // todo for what purpose?
        $this->updateCounterLimits();
        $this->_updateProfileLink();
        foreach ($this->tenant->users as $userResource) {
            if ($userResource->aps->status === Service_APSApi_Api::APS_READY) {
                $user = new users();
                $user->fillResource($userResource->aps->id);
                $user->getUser();
            }
        }

        return [
            static::TENANT => $this->fetchTenantLimitsAndUsages(),
            static::USER_USAGE => $this->getUserUsages(),
            static::ERROR => null,
        ];
    }

    public function createAdministrator($parameters, $tenantId = '', $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrators()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function updateAdministrator($adminId, $parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function deleteAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getUserGuideUrl()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministratorConsole($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function resendAdministratorLink($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function enableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function disableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function adminUrlRedirect()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getDatacenters()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlans()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function setStorageForPlan($planId, $parameters)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function unsetStorageForPlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    /* ======= OTHER METHODS ======= */

    protected function createDefaultUserResource($tenant, $serviceUser)
    {
        $user = \APS\TypeLibrary::newResourceByTypeId(\APS\TypeLibrary::getTypeByServiceId('users'));
        $user->user->set($serviceUser);
        $user->tenant->set($tenant);
        $user->privileged = users::ADMINISTRATOR_PRIVILEGES_YES;
        $user->activateDuringCreation = false;
        $this->initUserFormProfile($user);

        return $user;
    }

    /**
     * @param $serviceUser
     */
    protected function provisionDefaultUser($serviceUser)
    {
        $tenant = $this->tenant;

        $user = $this->createDefaultUserResource($tenant, $serviceUser);

        $apsc = $this->getApsc()->impersonate($tenant);
        $apsc->resetSession();
        $apsc->provisionResource($user);
    }

    /**
     * @return object
     */
    protected function getApplicationConfig()
    {
        return Settings::getInstance()->getApplicationConfig();
    }

    public function resolveDatacenterId()
    {
        $tenant = $this->tenant;
        // Do nothing if datacenter ID already was set from external side (online store / migration script)
        if (isset($tenant->datacenterId)) {
            return $tenant->datacenterId;
        }

        $datacenters = $tenant->global->datacenters;
        if ($datacenters->count() === 0) {
            throw new \Rest\RestException(401, static::EXCEPTION_NEED_TO_SET_CREDENTIALS_MESSAGE);
        }

        $profile = $tenant->profile;
        $storage = $profile->storagePolicy == profile::STORAGE_POLICY_PARENT
            ? $this->resolveProviderStorage($profile)
            : $this->resolveResellerStorage();

        return $storage->datacenterId;
    }

    public function resolveBackupStorageId()
    {
        $tenant = $this->tenant;

        $profile = $tenant->profile;
        $storage = $profile->storagePolicy == profile::STORAGE_POLICY_PARENT
            ? $this->resolveProviderStorage($profile)
            : $this->resolveResellerStorage();

        if ($tenant->datacenterId != $storage->datacenterId) {
            throw new \Rest\RestException(409, Str::format(
                'Please select storage from datacenter "%s" in the Service Profile "%s".',
                $tenant->datacenterId, $profile->resourceTypeId
            ));
        };

        return $storage->storageId;
    }

    /**
     * @return array
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function getUserUsages()
    {
        $canManageUserQuota = $this->canManageUserQuota();
        $offeringItemsToCounters = Arr::map(
            $this->fetchSubscriptionCounters(),
            function ($counter) {
                /** @var ApplicationCounter $counter */
                return $counter->getOfferingItemName();
            },
            function ($counter) {
                /** @var ApplicationCounter $counter */
                return $counter->getName();
            },
            false
        );
        $tenant = $this->tenant;
        // todo check why it is not array of users, but list of user ids
        /** @var string[] $userIds */
        $userIds = $this->getCloudApi()->getTenantUsers($tenant->tenantId);
        $usersUsages = [];

        foreach ($userIds as $userId) {
            $usageItems = [];
            $user = $this->getCloudApi()->getUser($userId);
            $userUsages = $this->getCloudApi()->getTenantUsages(array_first($user->getPersonalTenantId()));
            foreach ($userUsages as $usageItem) {
                $offeringItemName = $usageItem->getName();
                if (!isset($offeringItemsToCounters[$offeringItemName])) {
                    continue;
                }

                $counterName = $offeringItemsToCounters[$offeringItemName];
                $offeringItem = $usageItem->getOfferingItem();
                $offeringItemQuotaValue = $offeringItem->hasQuota() && $offeringItem->getQuota()->hasValue()
                    ? $offeringItem->getQuota()->getValue()
                    : self::UNLIMITED_VALUE;

                $usage = $usageItem->hasValue() ? $usageItem->getValue() : 0;
                $limit = $offeringItem->getStatus() === ApiInterface::OFFERING_ITEM_STATUS_ACTIVE ? $offeringItemQuotaValue : 0;
                $limit = $this->getCounterLimit(
                    $canManageUserQuota,
                    Arr::getNestedItem($tenant, [$counterName, 'limit'], null),
                    $limit);
                $usageItems[$counterName] = [
                    'usage' => $usage,
                    'limit' => $limit,
                ];
            }
            $usersUsages[$userId] = $usageItems;
        }

        return $usersUsages;
    }

    /**
     * @param $profile
     * @return object
     * @throws \Rest\RestException
     */
    private function resolveProviderStorage($profile)
    {
        $tenant = $this->tenant;
        $storages = is_array($profile->storages) ? $profile->storages : [];
        $storagesCount = count($storages);
        if ($storagesCount === 0) {
            throw new \Rest\RestException(400, self::EXCEPTION_SELECT_STORAGE_IN_BACKUP_PROFILE_MESSAGE);
        }

        if ($storagesCount === 1) {
            return reset($storages);
        }

        $subscriptionParametersManager = $this->getSubscriptionParametersManager();
        $storageOptionId = $subscriptionParametersManager->getParameterStorageId($tenant->subscription->subscriptionId);

        foreach ($storages as $storage) {
            if ($storageOptionId != datacenter::calculateStorageOptionId($storage->datacenterId, $storage->storageId)) {
                continue;
            }

            return $storage;
        }

        throw new \Rest\RestException(400, self::EXCEPTION_UNKNOWN_STORAGE_MESSAGE);
    }

    private function resolveResellerStorage()
    {
        $tenant = $this->tenant;
        $parentTenant = $this->_getParentTenant($tenant);
        if (is_null($parentTenant)) {
            return null;
        }
        $planId = $this->_getServicePlanId();
        $storage = $this->_getStorageForPlan($parentTenant, $planId);

        if (is_null($storage)) {
            throw new Exception(self::EXCEPTION_RESELLER_HAS_TO_SET_STORAGE_FOR_PLAN_MESSAGE);
        }

        return (object)[
            'datacenterId' => $storage[self::FIELD_STORAGE_DATACENTER_ID],
            'storageId' => $storage[self::FIELD_STORAGE_ID],
        ];
    }

    /**
     * @param $tenant
     * @return \APS\ResourceBase
     */
    private function _getParentTenant($tenant)
    {
        $accountId = $this->_getParentAccountId($tenant->accountId);
        $tenants = $this->getApsApi()->getTenantsByAccountId($accountId);
        return count($tenants) ? $tenants[0] : null;
    }

    /**
     * @param int $accountId
     * @return int
     */
    private function _getParentAccountId($accountId)
    {
        $openApi = $this->getOpenApi();
        $accountInfo = $openApi->getAccountInfo($accountId);
        return $accountInfo->getParentAccountId();
    }

    /**
     * @return int
     * @throws Exception | Service_PbaApi_Api_Exception
     */
    private function _getServicePlanId()
    {
        $tenant = $this->tenant;
        $pbaApi = $this->getPbaApi();
        $subscriptionDetails = $pbaApi->getSubscriptionDetails($tenant->subscription->subscriptionId);
        return $subscriptionDetails->getPlanId();
    }

    /**
     * @throws \Acronis\App\UsageReport\Exception\ManagerException
     * @throws \Rest\RestException
     */
    private function _retrieveUsageForCustomer()
    {
        $this->updateCountersUsage($this->tenant->datacenter);
    }

    public static function getCustomerFolderName($customerName)
    {
        $customerName = mb_strtoupper($customerName);
        $customerName = preg_replace('/[^A-Z\s]/', '', $customerName);
        $customerName = preg_replace('/\s+/', ' ', $customerName);
        $customerName = trim($customerName);
        if (strlen($customerName) === 1) {
            return $customerName . 'A';
        }
        $words = explode(' ', $customerName);
        if (count($words) > 1) {
            $folder = $words[0][0] . $words[1][0];
        } elseif (count($words) > 0 && strlen($words[0]) > 1) {
            $folder = $words[0][0] . $words[0][1];
        } else {
            $folder = static::DEFAULT_FOLDER_NAME;
        }

        return $folder;
    }

    /**
     * @param $parentMspGroupId
     * @return int|string
     * @throws Exception
     */
    private function getOrCreateFolderGroupId($parentMspGroupId)
    {
        try {
            $folderName = static::getCustomerFolderName($this->tenant->account->companyName);
            $parentId = $this->convertToTenantId($parentMspGroupId, $this->getMspApi());
            $groupId = $this->folderNameToParentIdCache->get(new FolderKey($parentId, $folderName));
            if (!is_null($groupId)) {
                return $groupId;
            }
            $groupId = $this->getMspGroupManager()->createOrFindFolder($this->tenant, $parentMspGroupId, $folderName);
            $this->folderNameToParentIdCache->set(new FolderKey($parentId, $folderName), $groupId);

            return $groupId;
        } catch (Exception $e) {
            \Log::getInstance()->error('Error during folder creation / get: "%s".', $e->getMessage());
            throw $e;
        }
    }

    /**
     * @return void
     */
    private function _updateGroupForCustomer()
    {
        $tenant = $this->tenant;
        $mspGroupManager = $this->getMspGroupManager();
        $newParentGroupId = null;
        $mspApi = $this->getMspApi();
        if ($this->foldersModeEnabled()) {
            $leafResellerMspGroupId = $this->_getOrCreateParentGroupId();
            $folderName = static::getCustomerFolderName($tenant->account->companyName);
            $cache = new FolderCache();
            $newParentGroupId = $cache->get(new FolderKey($this->convertToTenantId($leafResellerMspGroupId, $mspApi), $folderName));
            if (is_null($newParentGroupId)) {
                $newParentGroupId = $mspGroupManager->createOrFindFolder($this->tenant, $leafResellerMspGroupId, $folderName);
                $cache->set(new FolderKey($this->convertToTenantId($leafResellerMspGroupId, $mspApi), $folderName), $newParentGroupId);
            }
        }
        try {
            $mspGroupManager->updateGroupForCustomer($tenant, $newParentGroupId);
        } catch (\Exception $e) {
            if ($this->foldersModeEnabled()) {
                // for performance, we avoid assigning Folder offering items unless necessary
                $newParentGroup = $mspApi->getGroupById($newParentGroupId);
                $newParentGroupUuid = $newParentGroup->getUuid();
                $mspGroupManager->updateFolderOfferingItems($newParentGroupUuid, $tenant, $this);
                $mspGroupManager->updateGroupForCustomer($tenant, $newParentGroupId);
            }
        }
    }

    /**
     * @return int
     */
    private function _createGroupForCustomer()
    {
        \Log::getInstance()->debug(__METHOD__);
        $tenant = $this->tenant;
        $parentMspGroupId = $this->_getOrCreateParentGroupId();
        $mspGroupManager = $this->getMspGroupManager();
        $mspApi = $this->getMspApi();
        if ($this->foldersModeEnabled()) {
            $parentMspGroupId = $this->getOrCreateFolderGroupId($parentMspGroupId);
        }
        try {
            $groupId = $mspGroupManager->createGroupForCustomer($tenant, $parentMspGroupId);
        } catch (\Exception $e) {
            if ($this->foldersModeEnabled()) {
                // for performance, we avoid assigning Folder offering items unless necessary
                $newParentGroup = $mspApi->getGroupById($parentMspGroupId);
                $parentMspGroupUuid = $newParentGroup->getUuid();
                $mspGroupManager->updateFolderOfferingItems($parentMspGroupUuid, $tenant, $this);
                $groupId = $mspGroupManager->createGroupForCustomer($tenant, $parentMspGroupId);
            }
        }
        return $groupId;
    }


    private function obtainUsersLimitsFromCloud(?string $editionSwitchedTo, $usersOis)
    {
        $tenant = $this->tenant;
        $isNoManageMode = !$this->canManageUserQuota();

        foreach ($tenant->users as $user) {
            try {
                $newUserOfferingItems = [];
                foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
                    $applicationId = $this->getCloudApi()->getApplicationByType($applicationType)->getId();
                    $userResource = new \users();
                    $userResource->fillResource($user->aps->id);
                    $userOis =
                        ($usersOis !== null
                            && array_key_exists($applicationId, $usersOis)
                            && array_key_exists($userResource->lazyGetPersonalTenantId(), $usersOis[$applicationId])) ?
                        $usersOis[$applicationId][$userResource->lazyGetPersonalTenantId()] : null;
                    $newOfferingItems = $this->obtainUserLimitsFromCloud($userResource, $applicationId, $isNoManageMode, $editionSwitchedTo, $userOis);
                    $newUserOfferingItems = array_merge($newUserOfferingItems, $newOfferingItems);
                }

                $user->offeringItems = $newUserOfferingItems;
                $this->getApsc()->updateResource($user->resource);

                try {
                    $this->updateUserCloudAccount($user);
                } catch (Service_BaseClient_HTTPException $e) {
                    if ($e->getCode() !== Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                        throw $e;
                    }
                    $this->updateUserCloudAccount($user);
                }
            } catch (CloudDoesNotMatchWithOsaException $e) {
                \Log::getInstance()->warning("Skip user update. User's ({$user->userId}) offering items in the cloud does not match to subscription's resources in OSA. " . $e->getMessage());
            }
        }
    }

    /**
     * @param $user
     * @throws Service_BaseClient_HTTPException
     */
    private function updateUserCloudAccount($user)
    {
        try {
            $userManager = $this->createUserManagerInstance($this->getMspApi());
            foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
                $userManager->updateUserPersonalQuotas($user, $applicationType);
            }
        } catch (Service_BaseClient_HTTPException $e) {
            $skipCodes = [
                Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_FOUND,
                Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_AVAILABLE,
            ];
            if (!in_array($e->getCode(), $skipCodes)) {
                throw $e;
            }
            Log::getInstance()->warning(
                'Unable to update user "%s". Error: "%s".',
                $user->mspUserId, $e->getMessage()
            );
        }
    }

    /**
     * @return bool
     */
    private function canManageUserQuota()
    {
        $tenant = $this->tenant;
        $profile = $tenant->profile;
        if ($profile->isEmpty()) {
            $userQuota = true;
        } else {
            $userQuota = $profile->userQuota !== Service_APSApi_Api::PROFILE_USER_QUOTA_OFF;
        }
        return $userQuota;
    }

    /**
     * @throws Service_BaseClient_HTTPException
     */
    private function _deleteGroupForCustomer()
    {
        $cloudApi = $this->getCloudApi();

        $tenantInfo = $cloudApi->getTenant($this->tenant->tenantId);

        if (is_null($tenantInfo)) {
            Log::getInstance()->warning(sprintf(self::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $this->tenant->tenantId));
            throw new \Exception(sprintf(self::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $this->tenant->tenantId));
        }

        $version = $tenantInfo->getVersion();

        if ($tenantInfo->getEnabled()) {
            $tenantInfo = $cloudApi->disableTenant($tenantInfo->getId(), $version);
            $version = $tenantInfo->getVersion();
        }
        $cloudApi->deleteTenant($tenantInfo->getId(), $version);
    }

    private function getCounterLimit($canManageUserQuotas, $tenantCounterLimit, $userCounterLimit)
    {
        if ($canManageUserQuotas || $tenantCounterLimit) {
            return $userCounterLimit;
        }

        return $tenantCounterLimit;
    }

    private function _upgradeUsers()
    {
        $tenant = $this->tenant;
        $mspApi = $this->getMspApi();
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        $status = $groupInfo->getStatus();
        $version = $groupInfo->getVersion();
        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $version = $this->getCloudApi()->enableTenant($tenant->tenantId, $version)->getVersion();
        }

        $exception = null;
        try {
            $users = $tenant->users;
            if (!count($users)) {
                $users = $this->getApsApi()->getUsersForTenant($tenant);
            }
            foreach ($users as $resource) {
                $this->upgradeUser($resource->aps->id);
            }
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            $exception = $e;
        }

        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $this->getCloudApi()->disableTenant($tenant->tenantId, $version);
        }

        if (!is_null($exception)) {
            throw $exception;
        }
    }

    /**
     * @param $userId
     */
    private function upgradeUser($userId)
    {
        $user = $this->tenant->createUser();
        $user->fillResource($userId);
        $user->upgrade();
    }

    /**
     * @param $mspGroupId
     * @param $endUserLogin
     * @return null|Service_MspApiClient_Response_Struct_UserItem
     * @throws Exception
     */
    private function findUserByLogin($mspGroupId, $endUserLogin)
    {
        $users = $this->getMspApi()->getUsers($mspGroupId);
        foreach ($users as $user) {
            if ($user->getLogin() === $endUserLogin) {
                return $user;
            }
        }
        return null;
    }


    /**
     * The algorithm of this method is described in
     * https://adn.acronis.com/display/APS/User+quota+calculation+algorithm
     *
     * @param \users $user
     * @param string $applicationId
     * @param bool $isNoManageMode
     * @param string|null $editionSwitchedTo
     * @return array
     * @throws CloudDoesNotMatchWithOsaException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function obtainUserLimitsFromCloud($user, $applicationId, $isNoManageMode, ?string $editionSwitchedTo, $userOis)
    {
        $userOfferingItemsFromCloud = $userOis ?? $this->getCloudApi()->getTenantApplicationOfferingItems($user->lazyGetPersonalTenantId(), $applicationId);
        if (!$userOfferingItemsFromCloud) {
            return [];
        }

        $newUserOfferingItems = [];
        if ($isNoManageMode) {
            foreach ($userOfferingItemsFromCloud as $userOfferingItemFromCloud) {
                $newUserOfferingItems[] =
                    new UserOfferingItem($userOfferingItemFromCloud->getName(), self::UNLIMITED_VALUE);
            }
            return $newUserOfferingItems;
        }

        $existentUserOfferingItemsLimits = [];
        if ($user->offeringItems !== null) {
            /** @var UserOfferingItem $existentOfferingItem */
            foreach ($user->offeringItems as $existentOfferingItem) {
                $existentUserOfferingItemsLimits[$existentOfferingItem->name] = $existentOfferingItem->limit;
            }
        }

        //Steps 1-4
        $turnedOffUnlimitedUsageTypes = $this->getTurnedOffUnlimitedUsageTypes($user);

        $this->checkCloudVsOsaResources($user, $userOfferingItemsFromCloud);

        // TODO: Merge nomange and and non available for user. Store to APS only manageable user OIs.
        foreach ($userOfferingItemsFromCloud as $userOfferingItemFromCloud) {
            $offeringItemName = $userOfferingItemFromCloud->getName();
            if ($userOfferingItemFromCloud->hasQuota() && $userOfferingItemFromCloud->getQuota()->hasValue()) {
                // Step 6
                $newLimit = $userOfferingItemFromCloud->getQuota()->getValue();
            } elseif (!$this->isOfferingItemAvailableForUser($offeringItemName)) {
                // TODO: check why we do not add this OI to output array in previous implementation
                continue;
            } else {
                if (array_key_exists($offeringItemName, $existentUserOfferingItemsLimits)) {
                    // Step 7a
                    $newLimit = $existentUserOfferingItemsLimits[$offeringItemName];
                } else {
                    // Step 7b
                    $counter = $this->getCountersManager()->getCounterByOfferingItemName($offeringItemName);
                    $usageType = $this->sanitizeUsageType($counter->getUsageType());
                    $newLimit =
                        (in_array($usageType, $turnedOffUnlimitedUsageTypes)
                            || ($editionSwitchedTo != '' && $editionSwitchedTo === $userOfferingItemFromCloud->getEdition())
                        ) ?
                        self::UNLIMITED_VALUE : 0;
                }
            }
            $newUserOfferingItems[] = new UserOfferingItem($offeringItemName, $newLimit);
        }
        return $newUserOfferingItems;
    }

    /**
     * @param users $user
     * @return array
     */
    protected function getTurnedOffUnlimitedUsageTypes($user)
    {
        $existentUnlimitedOfferingItemsNames = $this->getExistentUnlimitedOfferingItems($user->offeringItems);
        $newOfferingItemsNames = $this->calculateOfferingItemsEnabledByOsaResources($user->tenant->subscription->resources());
        $turnedOffOfferingItemsNames = array_diff($existentUnlimitedOfferingItemsNames, $newOfferingItemsNames);
        $unlimitedCounters = $this->getCountersManager()->findCountersByOfferingItemNames($turnedOffOfferingItemsNames);
        $turnedOffUnlimitedUsageTypes = [];
        foreach ($unlimitedCounters as $counter) {
            if (!str_contains($counter->getOfferingItemName(), 'secondary')) {
                Log::getInstance()->debug('getTurnedOffUnlimitedUsageTypes ' . $counter->getName() . ' ' . $counter->getUsageType() . ' ==> ' . $this->sanitizeUsageType($counter->getUsageType()));
                $turnedOffUnlimitedUsageTypes[] = $this->sanitizeUsageType($counter->getUsageType());
            }
        }
        return array_unique($turnedOffUnlimitedUsageTypes);
    }

    private function sanitizeUsageType(string $usageType): string
    {
        $prefixes = [
            'pack_adv_backup_',
            'pack_adv_'
        ];

        return str_replace($prefixes, '', $usageType);

    }

    /**
     * @param array $offeringItems
     * @return array
     */
    protected function getExistentUnlimitedOfferingItems($offeringItems)
    {
        if ($offeringItems === null) {
            return [];
        }
        $names = [];
        /** @var UserOfferingItem $offeringItem */
        foreach ($offeringItems as $offeringItem) {
            if ($offeringItem->limit == self::UNLIMITED_VALUE) {
                $names[] = $offeringItem->name;
            }
        }
        return $names;
    }

    /**
     * @param array $resources
     * @return array
     */
    protected function calculateOfferingItemsEnabledByOsaResources($resources)
    {
        $names = [];
        $counterManager = $counter = $this->getCountersManager();
        foreach ($resources as $resource) {
            if (!isset($resource->property)) {
                continue;
            }
            $names[] = $counterManager->getCounterByName($resource->property)->getOfferingItemName();
        }
        return $names;
    }

    /**
     * @param $offeringItemName
     * @return bool
     */
    protected function isOfferingItemAvailableForUser($offeringItemName)
    {
        // TODO This is dirty hack in order not to refactor logic which operates with not mutually ambiguous counters <-> offering items
        // TODO: Refactoring: exclude Arr::map constructions and design it inside CountersManager (interface)
        $offeringItemsToCountersNames = $this->memoize(function() {
            $names = [];
            foreach ($this->getCountersManager()->getAllCounters(CountersManager::FILTER_EXCLUDE_NON_USER_COUNTERS) as $counter) {
                $names[$counter->getOfferingItemName()] = null;
            }
            return $names;
        });

        return array_key_exists($offeringItemName, $offeringItemsToCountersNames);
    }

    /**
     * @return mixed
     */
    private function foldersModeEnabled()
    {
        return $this->getConfig()->getCloudApiConfig()[ConfigSchema::NODE_CLOUD_API_SPLIT_BY_FOLDERS];
    }

    /**
     * @param users $user
     * @param array $userOfferingItemsFromCloud
     * @throws CloudDoesNotMatchWithOsaException
     */
    private function checkCloudVsOsaResources($user, array $userOfferingItemsFromCloud)
    {
        $offeringItemsFromOsaResources = $this->calculateOfferingItemsEnabledByOsaResources($user->tenant->subscription->resources());
        $offeringItemsFromOsaResources = array_flip($offeringItemsFromOsaResources);
        foreach ($userOfferingItemsFromCloud as $cloudOfferingItem) {
            $cloudOfferingItemName = $cloudOfferingItem->getName();
            if (!array_key_exists($cloudOfferingItemName, $offeringItemsFromOsaResources)) {
                throw new CloudDoesNotMatchWithOsaException("Cloud offering \"{$cloudOfferingItemName}\" does not match to the subscription resources");
            }
        }
    }
}
