<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\Common\CodeGenerator;

use Acronis\Common\CodeGenerator\Exception\CanNotRollbackFileException;
use Acronis\Common\CodeGenerator\Exception\InterpolatorNotSetException;
use Acronis\Common\Utils\Arr;
use Acronis\Common\Utils\CodegenHelper;
use Acronis\Common\Utils\Str;

class CodeGenerator
{
    /** @var InterpolatorInterface */
    private $interpolator;

    private $generatedCode = '';

    private $generatedJsonItems = [];

    /** @var CodegenHelper */
    private $sourceHelper;

    /**
     * CodeGenerator constructor.
     * @param CodegenHelper $sourceHelper
     * @param ?InterpolatorInterface $interpolator
     */
    public function __construct(CodegenHelper $sourceHelper, ?InterpolatorInterface $interpolator = null)
    {
        $this->sourceHelper = $sourceHelper;
        $this->interpolator = $interpolator;
    }

    /**
     * @param $contextItems
     * @param bool $testMode
     */
    public function generate(array $contextItems, bool $testMode = false)
    {
        foreach ($this->sourceHelper->getTemplateMappings() as $mappingName => $mappingOptions) {
            $filePath = $this->sourceHelper->getCodePath() . ($mappingOptions['relative_path'] ?? '');
            if ($testMode) {
                $lineToAppend = ($mappingOptions['comment'] ?? '//') . ' ' . $mappingName . PHP_EOL;
                $dirPath = dirname($filePath);
                if (!is_dir($dirPath)) {
                    mkdir($dirPath, 0777, true);
                }
                if (!file_exists($filePath)) {
                    file_put_contents($filePath, $lineToAppend);
                } else if ($testMode) {
                    file_put_contents($filePath, $lineToAppend, FILE_APPEND | LOCK_EX);
                }
            } else if (!file_exists($filePath)) {
                echo sprintf('file "%s" not found, skipping.' . PHP_EOL, $filePath);
                continue;
            }

            $isJson = $mappingOptions['type'] == 'json';
            $trim = $mappingOptions['trim'] ?? true;
            $assoc = $mappingOptions['assoc'] ?? false;

            if (!empty($mappingOptions['custom_handler'])) {
                $this->sourceHelper->initGeneratorHandler($mappingOptions['custom_handler']);
            } else {
                $this->sourceHelper->resetHandler();
            }

            $this->clearGeneratedItems();
            if (!$testMode) {
                $this->sourceHelper->backupOriginalFile($filePath);
            }

            $this->customHandler($mappingOptions, 'on_load_custom_handler');


            foreach ($contextItems as $item) {
                if ($this->itemIsSkipped($item, $mappingOptions['skip_when'] ?? null)) {
                    continue;
                }
                $templateFilePath = $this->sourceHelper->getTemplatesPath() . '/' . $mappingOptions['template'];
                $renderedTemplate = $this->renderTemplate(file_get_contents($templateFilePath), $item, $item === end($contextItems));

                $item = $this->customHandler($mappingOptions, 'prepare_item_handler', $item);

                if ($isJson) {

                    $itemPaths = $this->parsePaths($mappingOptions['path'], $item);
                    $this->appendGeneratedJson($renderedTemplate, $itemPaths, $assoc);
                } else {
                    $this->appendGeneratedCode($renderedTemplate, $trim);
                }
            }
            $contents = $this->applyGeneratedItems(file_get_contents($filePath), $isJson, $mappingName);

            $this->sourceHelper->writeChanges($filePath, $contents);
            $this->sourceHelper->showGeneratedItems($isJson, $this->generatedJsonItems, $this->generatedCode);
        }
    }

    /**
     * @param $mappingOptions
     * @param $handler
     * @param null $data
     * @return mixed|null
     */

    private function customHandler(array $mappingOptions, $handlerName, $data = null)
    {
        if (!empty($mappingOptions[$handlerName])) {
            $handlerObject = $this->sourceHelper->getHandler();
            if ($handlerObject && method_exists($handlerObject, $mappingOptions[$handlerName])) {
                return $handlerObject->{$mappingOptions[$handlerName]}($mappingOptions, $data);
            }
        }
        return $data;
    }

    /**
     * @throws CanNotRollbackFileException
     */
    public function rollback()
    {
        foreach ($this->sourceHelper->getTemplateMappings() as $mappingName => $mappingOptions) {
            // it should be fixed (and all path transformations should be in a single place)
            $filePath = $this->sourceHelper->getCodePath() . '/' . ltrim($mappingOptions['relative_path'], '/');
            $this->sourceHelper->rollback($filePath);
        }
    }

    private function clearGeneratedItems()
    {
        $this->generatedCode = '';
        $this->generatedJsonItems = [];
    }

    /**
     * @param $template
     * @param $contextItem
     * @param bool $isLastItem
     * @return mixed
     */
    private function renderTemplate($template, $contextItem, $isLastItem = false)
    {
        if ($this->interpolator === null) {
            throw new InterpolatorNotSetException('Interpolator is not set, template can not be rendered.');
        }

        return $this->interpolator->interpolate($template, $contextItem, $isLastItem ? ['is_last_item' => true] : []);
    }

    /**
     * Replace path tokens(if any exist) with corresponding item values
     *
     * Example:
     *  $path = 'a.{b}.{c}'
     *  $item = [
     *    'b' => ['b1'],
     *    'c' => ['c1', 'c2'],
     *  ]
     *
     *  return = [
     *    'a.b1.c1',
     *    'a.b1.c2'
     *  ]
     *
     * @param $path
     * @param $item
     * @return array
     */
    private function parsePaths($path, $item)
    {
        if (!preg_match_all("/{(\S+?)}/", $path, $pathProperties)) {
            return [$path];
        }
        $pathProperties = $pathProperties[1];

        $pathReplacements = [[]];
        foreach ($pathProperties as $pathProperty) {
            if (!isset($item[$pathProperty])) {
                // paths cannot be parsed for this item
                return [];
            }
            $pathValues = $item[$pathProperty];
            $pathReplacements = $this->addPathReplacements($pathReplacements, $pathProperty, $pathValues);
        }

        return array_map(function ($replacements) use ($path) {
            $replaceKeys = array_keys($replacements);
            $replaceValues = array_values($replacements);
            return str_replace($replaceKeys, $replaceValues, $path);
        }, $pathReplacements);
    }

    private function addPathReplacements($currentReplacements, $key, $values)
    {
        $replacements = [];
        foreach ($values as $value) {
            $newReplacements = array_map(function ($replacement) use ($key, $value) {
                $replacementKey = '{' . $key . '}';
                $replacement[$replacementKey] = $value;
                return $replacement;
            }, $currentReplacements);
            $replacements = array_merge($replacements, $newReplacements);
        }

        return $replacements;
    }

    private function appendGeneratedJson($item, $paths = [], $assoc = false)
    {
        $item = json_decode($item, true);
        foreach ($paths as $path) {
            if (!isset($this->generatedJsonItems[$path])) {
                $this->generatedJsonItems[$path] = [];
            }
            if ($assoc) {
                $this->generatedJsonItems[$path] = array_merge($this->generatedJsonItems[$path], $item);
            } else {
                $this->generatedJsonItems[$path][] = $item;
            }
        }
    }

    private function appendGeneratedCode($item, $trim = true)
    {
        if (trim($item) != '') {
            $this->generatedCode .= ($trim ? trim($item) : $item) . PHP_EOL;
        }
    }

    private function applyGeneratedItems($content, $isJson, $mappingName)
    {
        if ($isJson) {
            $object = json_decode($content, true);
            foreach ($this->generatedJsonItems as $path => $jsonItems) {
                Arr::appendToPath(
                    $object,
                    explode('.', $path),
                    $jsonItems
                );
            }
            return json_encode($object, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
        } else {
            return Str::appendBeforePlaceholder($content, $mappingName, $this->generatedCode);
        }
    }

    private function itemIsSkipped($item, $skipWhenTemplate)
    {
        if ($skipWhenTemplate !== null &&
            $this->interpolator->interpolate($skipWhenTemplate, $item, []) == 'true') {
            return true;
        }

        return false;
    }
}