<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\Common\Utils;

use Acronis\Common\CodeGenerator\Exception\CanNotRollbackFileException;
use Acronis\Common\Utils\Exception\UnknownSourceExtension;
use Symfony\Component\Yaml\Yaml;

class CodegenHelper
{
    private $savedFiles = [];

    private $templatesPath;

    private $codePath;

    private $dryRun;

    private $handler;

    private $testMappingName;

    /**
     * TemplatesHelper constructor.
     * @param $templatesPath
     * @param $dryRun
     * @param $testMappingName
     */
    public function __construct($templatesPath, $codePath, $dryRun, $testMappingName)
    {
        $this->templatesPath = rtrim($templatesPath, '/') . '/';
        $this->codePath = rtrim($codePath, '/') . '/';
        $this->dryRun = $dryRun;
        $this->testMappingName = $testMappingName;
    }

    /**
     * @param string $handler
     * @return mixed|null
     */
    public function initGeneratorHandler(string $handler)
    {
        $handlerPath = $this->templatesPath . $handler . '.php';

        if (file_exists($handlerPath)) {
            include_once($handlerPath);
            $this->handler = new $handler($this);
        } else {
            $this->handler = null;
        }
    }

    /**
     * @return void
     */
    public function resetHandler()
    {
        $this->handler = null;
    }

    /**
     * @return mixed
     */

    public function getHandler()
    {
        return $this->handler;
    }

    public static function loadItemsFromPath($filePath)
    {
        $extension = pathinfo($filePath, PATHINFO_EXTENSION);
        if ($extension === 'csv') {
            $handler = fopen($filePath, 'r');
            $rows = new CsvFileIterator($handler, true);
            $columns = [];
            foreach ($rows as $i => $row) {
                if ($i === 0) {
                    $columns = $row;
                } else if (count($row) >= count($columns)) { // for correct handling of last line and all not filled rows
                    $items[] = array_combine($columns, array_slice($row, 0, count($columns)));
                } else {
                    // log
                }
            }
        } else if (in_array($extension, ['yml', 'yaml'])) {
            $items = Yaml::parseFile($filePath);
        } else {
            throw new UnknownSourceExtension($extension);
        }

        return $items;
    }

    /**
     * @param $filePath
     * @throws CanNotRollbackFileException
     */
    public function rollback($filePath)
    {
        $backupFilePath = realpath($filePath . '.orig');
        $filePath = realpath($filePath);
        if (file_exists($filePath) && file_exists($backupFilePath)) {
            copy($backupFilePath, $filePath);
        } else {
            throw new CanNotRollbackFileException($backupFilePath, $filePath);
        }
    }

    public function backupOriginalFile($filePath)
    {
        if (!$this->dryRun && !isset($this->savedFiles[$filePath])) {
            $this->savedFiles[$filePath] = true;
            copy($filePath, $filePath . '.orig');
        }
    }

    public function getTemplateMappings()
    {
        $mappingsFilePath = $this->templatesPath . '/mappings.yaml';
        $mappings = Yaml::parseFile($mappingsFilePath);
        foreach ($mappings as $mappingName => $options) {
            if ($this->dryRun && $mappingName !== $this->testMappingName) {
                continue;
            }
            yield $mappingName => $options;
        }
    }

    public function getTemplatesPath()
    {
        return $this->templatesPath;
    }

    public function getCodePath()
    {
        return $this->codePath;
    }

    public function writeChanges($filePath, $contents)
    {
        if (!$this->dryRun) {
            file_put_contents($filePath, $contents);
        }
    }

    public function showGeneratedItems($isJson, $generatedItemsJson, $generatedCode)
    {
        if ($this->dryRun) {
            echo $isJson ? json_encode($generatedItemsJson) : $generatedCode;
        }
    }
}