<?php

use Acronis\Common\Utils\Arr;
use Acronis\Common\Utils\Exception\InvalidDestinationException;
use Acronis\Common\Utils\Exception\InvalidItemsException;

/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */
class ClassWithMagicMethodIsset
{
    public function __isset($property)
    {
    }
}

class ArrTest extends \PHPUnit\Framework\TestCase
{
    public function dataProvider()
    {
        $testTypes = [
            'string_empty' => '',
            'string' => 'string',
        ];

        $testFormats = [
            '' => "",
            'quotas1' => "''",
            'quotas2' => '""',
            'equals1' => '=',
            'equals2' => '1=2=3',
        ];

        return [
            [$testTypes],
            [$testFormats],
        ];
    }

    public function dataProviderHas()
    {
        $arrayAccess = $this->createMock(ArrayAccess::class);
        $arrayAccess->expects($this->once())
            ->method('offsetExists')
            ->with('key')
            ->willReturn(true);

        $arrayIsset = $this->createMock(ClassWithMagicMethodIsset::class, ['__isset']);
        $arrayIsset->expects($this->once())
            ->method('__isset')
            ->with('key')
            ->willReturn(true);

        $buildStdClass = function (array $properties = []) {
            $object = new stdClass();
            foreach ($properties as $property => $value) {
                $object->{$property} = $value;
            }

            return $object;
        };

        return [
            // 'test name' => [key, data, expected]

            'ArrayAccess' => ['key', $arrayAccess, true],

            '__isset' => ['key', $arrayIsset, true],

            'array false value' => ['key', ['key' => false], true],
            'array zero value' => ['key', ['key' => 0], true],
            'array null value' => ['key', ['key' => null], true],
            'array empty value' => ['key', ['key' => ''], true],
            'array key not exists' => ['key', ['another_key' => ''], false],
            'array empty' => ['key', [], false],

            'object false value' => ['key', (object)['key' => false], true],
            'object zero value' => ['key', (object)['key' => 0], true],
            'object null value' => ['key', (object)['key' => null], true],
            'object empty value' => ['key', (object)['key' => ''], true],
            'object key not exists' => ['key', (object)['another_key' => ''], false],
            'object empty' => ['key', (object)[], false],

            'stdClass false value' => ['key', $buildStdClass(['key' => false]), true],
            'stdClass zero value' => ['key', $buildStdClass(['key' => 0]), true],
            'stdClass null value' => ['key', $buildStdClass(['key' => null]), true],
            'stdClass empty value' => ['key', $buildStdClass(['key' => '']), true],
            'stdClass key not exists' => ['key', $buildStdClass(['another_key' => '']), false],
            'stdClass empty' => ['key', $buildStdClass([]), false],
        ];
    }

    public function dataProviderGetNested()
    {
        $arrayIsset = $this->createMock(ClassWithMagicMethodIsset::class, ['__isset'], [], '', false);
        $arrayIsset->expects($this->once())
            ->method('__isset')
            ->with('d_key')
            ->willReturn(true);
        $arrayIsset->d_key = 'bar';

        $arrayAccess = $this->createMock(ArrayAccess::class);
        $arrayAccess->expects($this->once())
            ->method('offsetExists')
            ->with('c_key')
            ->willReturn(true);
        $arrayAccess->expects($this->once())
            ->method('offsetGet')
            ->with('c_key')
            ->willReturn($arrayIsset);

        return [
            'default value contains specified path' => [['a_key', 'b_key'], ['a_key' => 'foo'], ['a_key' => 'b_key'], ['a_key' => 'b_key']],
            'depth 2, has nested property' => [['a_key', 'b_key'], ['a_key' => ['b_key' => 'bar']], 'baz', 'bar'],
            'depth 2, default value' => [['a_key', 'b_key'], ['a_key' => ['bbbb_key' => 'bar']], 'baz', 'baz'],
            'mixed data (object, array, array access, array isset)' => [['a_key', 'b_key', 'c_key', 'd_key'], (object)['a_key' => ['b_key' => $arrayAccess]], 'baz', 'bar']
        ];
    }

    public function dataProviderGet()
    {
        $buildArrayAccess = function (array $properties = [], array $notExistProperties = []) {
            $object = $this->createMock(ArrayAccess::class);
            foreach ($properties as $property => $value) {
                $object->expects($this->once())
                    ->method('offsetExists')
                    ->with($property)
                    ->willReturn(true);

                $object->expects($this->once())
                    ->method('offsetGet')
                    ->with($property)
                    ->willReturn($value);
            }

            foreach ($notExistProperties as $property) {
                $object->expects($this->once())
                    ->method('offsetExists')
                    ->with($property)
                    ->willReturn(false);
            }

            return $object;
        };

        $obj = new stdClass();

        return [

            'ArrayAccess zero value' => ['key', $buildArrayAccess(['key' => 0]), $this->anything(), 0],
            'ArrayAccess empty value' => ['key', $buildArrayAccess(['key' => '']), $this->anything(), ''],
            'ArrayAccess false value' => ['key', $buildArrayAccess(['key' => false]), $this->anything(), false],
            'ArrayAccess null value' => ['key', $buildArrayAccess(['key' => null]), $this->anything(), null],
            'ArrayAccess array value' => ['key', $buildArrayAccess(['key' => []]), $this->anything(), []],
            'ArrayAccess object value' => ['key', $buildArrayAccess(['key' => $obj]), $this->anything(), $obj],

            'ArrayAccess default zero value' => ['key', $buildArrayAccess([], ['key']), 0, 0],
            'ArrayAccess default empty value' => ['key', $buildArrayAccess([], ['key']), '', ''],
            'ArrayAccess default false value' => ['key', $buildArrayAccess([], ['key']), false, false],
            'ArrayAccess default null value' => ['key', $buildArrayAccess([], ['key']), null, null],
            'ArrayAccess default array value' => ['key', $buildArrayAccess([], ['key']), [], []],
            'ArrayAccess default object value' => ['key', $buildArrayAccess([], ['key']), $obj, $obj],

            'array zero value' => ['key', ['key' => 0], $this->anything(), 0],
            'array empty value' => ['key', ['key' => ''], $this->anything(), ''],
            'array false value' => ['key', ['key' => false], $this->anything(), false],
            'array null value' => ['key', ['key' => null], $this->anything(), null],
            'array array value' => ['key', ['key' => []], $this->anything(), []],
            'array object value' => ['key', ['key' => $obj], $this->anything(), $obj],

            'array default zero value' => ['key', [], 0, 0],
            'array default empty value' => ['key', [], '', ''],
            'array default false value' => ['key', [], false, false],
            'array default null value' => ['key', [], null, null],
            'array default array value' => ['key', [], [], []],
            'array default object value' => ['key', [], $obj, $obj],

            'object zero value' => ['key', (object)['key' => 0], $this->anything(), 0],
            'object empty value' => ['key', (object)['key' => ''], $this->anything(), ''],
            'object false value' => ['key', (object)['key' => false], $this->anything(), false],
            'object null value' => ['key', (object)['key' => null], $this->anything(), null],
            'object array value' => ['key', (object)['key' => []], $this->anything(), []],
            'object object value' => ['key', (object)['key' => $obj], $this->anything(), $obj],

            'object default zero value' => ['key', (object)[], 0, 0],
            'object default empty value' => ['key', (object)[], '', ''],
            'object default false value' => ['key', (object)[], false, false],
            'object default null value' => ['key', (object)[], null, null],
            'object default array value' => ['key', (object)[], [], []],
            'object default object value' => ['key', (object)[], $obj, $obj],
        ];
    }

    /**
     * @dataProvider dataProvider
     */
    public function testEncodeDecode($data)
    {
        $encodedData = Arr::encode($data);
        $this->assertEquals($data, Arr::decode($encodedData));
    }

    public function testDecode()
    {
        $this->assertEquals([], Arr::decode(null));
        $this->assertEquals([], Arr::decode(''));
        $this->assertEquals([
            'key1' => null,
        ],
            Arr::decode('key1')
        );
        $this->assertEquals([
            'key1' => null,
            'key2' => null,
        ],
            Arr::decode('key1' . Arr::PAIR_DELIMITER . 'key2')
        );
        $this->assertEquals([
            'key1' => '1',
        ],
            Arr::decode(
                'key1' . Arr::VALUE_DELIMITER . '1'
            )
        );
        $this->assertEquals([
            'key2' => '1',
        ],
            Arr::decode(
                'key2' . Arr::VALUE_DELIMITER . '1'
            )
        );
        $this->assertEquals([
            'key1' => '1',
            'key2' => '2',
        ],
            Arr::decode(
                'key1' . Arr::VALUE_DELIMITER . '1' . Arr::PAIR_DELIMITER .
                'key2' . Arr::VALUE_DELIMITER . '2'
            )
        );
        $this->assertEquals([
            'key1' => '1',
            'key2' => '2',
            'key3' => null,
        ],
            Arr::decode(
                'key1' . Arr::VALUE_DELIMITER . '1' . Arr::PAIR_DELIMITER .
                'key2' . Arr::VALUE_DELIMITER . '2' . Arr::PAIR_DELIMITER .
                'key3'
            )
        );
    }

    public function testEncode()
    {
        $this->assertEquals('', Arr::encode([]));
        $this->assertEquals(
            'key1',
            Arr::encode([
                'key1' => null,
            ])
        );
        $this->assertEquals(
            'key1' . Arr::PAIR_DELIMITER . 'key2',
            Arr::encode([
                'key1' => null,
                'key2' => null,
            ])
        );
        $this->assertEquals(
            'key1' . Arr::VALUE_DELIMITER . '1',
            Arr::encode([
                'key1' => '1',
            ])
        );
        $this->assertEquals(
            'key2' . Arr::VALUE_DELIMITER . '1',
            Arr::encode([
                'key2' => '1',
            ])
        );
        $this->assertEquals(
            'key1' . Arr::VALUE_DELIMITER . '1' . Arr::PAIR_DELIMITER .
            'key2' . Arr::VALUE_DELIMITER . '2',
            Arr::encode([
                'key1' => '1',
                'key2' => '2',
            ])
        );
        $this->assertEquals(
            'key1' . Arr::VALUE_DELIMITER . '1' . Arr::PAIR_DELIMITER .
            'key2' . Arr::VALUE_DELIMITER . '2' . Arr::PAIR_DELIMITER .
            'key3',
            Arr::encode([
                'key1' => '1',
                'key2' => '2',
                'key3' => null,
            ])
        );
    }

    /**
     * @dataProvider dataProviderHas
     */
    public function testHas($key, $data, $expected)
    {
        $this->assertEquals($expected, Arr::has($data, $key));
    }

    /**
     * @dataProvider dataProviderGet
     */
    public function testGet($key, $data, $defaultValue, $expected)
    {
        $this->assertEquals($expected, Arr::get($data, $key, $defaultValue));
    }

    /**
     * @dataProvider dataProviderGetNested
     */
    public function testGetNested($keys, $data, $defaultValue, $expected)
    {
        $this->assertEquals($expected, Arr::getNestedItem($data, $keys, $defaultValue));
    }

    public function testFindPlainValue()
    {
        $arr = ['a', 'b', 'c'];
        $findMe = 'b';
        $expected = 'b';
        $this->assertEquals($expected, Arr::find($arr, $findMe));
    }

    public function testFindByCallable()
    {
        $arr = ['1', '2', '3', 1, 2, 3];
        $findCall = function ($item) {
            return is_string($item) && $item === '2';
        };
        $expected = '2';
        $this->assertEquals($expected, Arr::find($arr, $findCall));
    }

    public function testMerge()
    {
        $limits = ['a' => 11, 'b' => 22, 'c' => 33, 'd' => 44,];
        $usages = ['a' => 1, 'b' => 2, 'c' => 3, 'e' => 4,];
        $expected = [
            'a' => ['limit' => 11, 'usage' => 1,],
            'b' => ['limit' => 22, 'usage' => 2,],
            'c' => ['limit' => 33, 'usage' => 3,],
            'd' => ['limit' => 44, 'usage' => 0,],
            'e' => ['usage' => 4,],
        ];

        $counters = Arr::merge([
            'limit' => $limits,
            'usage' => $usages,
        ], [
            'usage' => 0,
        ]);

        $this->assertEquals($expected, $counters);
    }

    public function testMap()
    {
        $countries = [
            ['country_code' => 'US', 'currency_code' => 'USD', 'name' => 'United States',],
            ['country_code' => 'RU', 'currency_code' => 'RUB', 'name' => 'Russia',],
            ['country_code' => 'FR', 'currency_code' => 'EUR', 'name' => 'France',],
        ];
        $expected = [
            'US' => 'USD',
            'RU' => 'RUB',
            'FR' => 'EUR',
        ];

        $currencies = Arr::map($countries, 'country_code', 'currency_code');

        $this->assertEquals($expected, $currencies);
    }

    public function testMapKeys()
    {
        $data = [
            'key1' => 'value1',
            'key2' => 'value2',
            'key3' => 'value3',
        ];
        $expected = [
            'value1' => 'key1',
            'value2' => 'key2',
            'value3' => 'key3',
        ];

        $result = Arr::map(
            $data,
            function ($value) {
                return $value;
            },
            function ($value, $key) {
                return $key;
            }
        );

        $this->assertEquals($expected, $result);
    }

    /**
     * @dataProvider jsonItemsDataProvider
     */
    public function testAppendJsonItemsToPath($sourceObject, $path, $items, $expectedObject, $expectedException = null)
    {
        if ($expectedException) {
            $this->expectException($expectedException);
        }

        Arr::appendToPath($sourceObject, $path, $items);

        if (!$expectedException) {
            $this->assertEquals($expectedObject, $sourceObject);
        }
    }

    public function jsonItemsDataProvider()
    {
        return [
            'Append items to existing path' => [
                'source_object' => [
                    'a' => [
                        'b' => [1, 2, 3]
                    ]
                ],
                'path' => ['a', 'b'],
                'items' => [4, 5, 6],
                'expected' => [
                    'a' => [
                        'b' => [1, 2, 3, 4, 5, 6]
                    ]
                ]
            ],
            'Append items to new path' => [
                'source_object' => [
                    'a' => [
                        'b' => [1, 2, 3]
                    ]
                ],
                'path' => ['a', 'c'],
                'items' => [4, 5, 6],
                'expected' => [
                    'a' => [
                        'b' => [1, 2, 3],
                        'c' => [4, 5, 6]
                    ]
                ]
            ],
            'Append items to existing path in mixed object' => [
                'source_object' => (object)[
                    'a' => [
                        'b' => [1, 2, 3]
                    ]
                ],
                'path' => ['a', 'b'],
                'items' => [4, 5, 6],
                'expected' => (object)[
                    'a' => [
                        'b' => [1, 2, 3, 4, 5, 6]
                    ]
                ]
            ],
            'Append invalid items' => [
                'source_object' => (object)[
                    'a' => [
                        'b' => [1, 2, 3]
                    ]
                ],
                'path' => ['a', 'b'],
                'items' => (object)[], // items to append are not an array
                'expected' => null,
                'expected_exception' => InvalidItemsException::class
            ],
            'Append items to invalid destination' => [
                'source_object' => (object)[
                    'a' => [
                        'b' => (object)[1, 2, 3] // destination path is not an array
                    ]
                ],
                'path' => ['a', 'b'],
                'items' => [4, 5, 6],
                'expected' => null,
                'expected_exception' => InvalidDestinationException::class
            ],
        ];
    }
}
